"""Jobs to create training data for ML potentials."""

import gzip
import logging
import os
import pickle
import shutil
import traceback
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import matplotlib.pyplot as plt
import numpy as np
from ase.constraints import voigt_6_to_full_3x3_stress
from ase.io import read, write
from atomate2.utils.path import strip_hostname
from emmet.core.math import Matrix3D
from jobflow.core.job import job
from phonopy.structure.cells import get_supercell
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.phonopy import get_phonopy_structure, get_pmg_structure
from pymatgen.io.vasp.outputs import Vasprun

from autoplex.data.common.utils import (
    ElementCollection,
    boltzhist_cur_dual_iter,
    boltzhist_cur_one_shot,
    create_soap_descriptor,
    cur_select,
    data_distillation,
    flatten,
    handle_rss_trajectory,
    mc_rattle,
    random_vary_angle,
    scale_cell,
    std_rattle,
    stratified_dataset_split,
    to_ase_trajectory,
)
from autoplex.fitting.common.regularization import set_custom_sigma

if TYPE_CHECKING:
    from ase import Atoms

logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")


@job
def convert_to_extxyz(job_output, pkl_file, config_type, factor):
    """
    Convert data and write extxyt file.

    Parameters
    ----------
    job_output:
        The (static) job output object.
    pkl_file: str
        A pickle file.
    config_type: str
        Configuration type of the data.
    factor: str
        String of a factor to resize cell parameters.

    """
    with open(Path(job_output.dir_name) / Path(pkl_file), "rb") as file:
        traj_obj = pickle.load(file)
    data = to_ase_trajectory(traj_obj=traj_obj)
    data[-1].write("tmp.xyz")
    file = read("tmp.xyz", index=":")
    for i in file:
        # TODO: enable switching to stress
        virial_list = -voigt_6_to_full_3x3_stress(i.get_stress()) * i.get_volume()
        i.info["REF_virial"] = " ".join(map(str, virial_list.flatten()))
        del i.calc.results["stress"]
        i.arrays["REF_forces"] = i.calc.results["forces"]
        del i.calc.results["forces"]
        i.info["REF_energy"] = i.calc.results["energy"]
        del i.calc.results["energy"]
        i.info["config_type"] = config_type
        i.pbc = True
    write("ref_" + factor + ".extxyz", file, append=True)

    return os.getcwd()


@job
def plot_force_distribution(
    cell_factor: float,
    path: str,
    x_min: int = 0,
    x_max: int = 5,
    bin_width: float = 0.125,
):
    """
    Plotter for the force distribution.

    Parameters
    ----------
    cell_factor: float
        Factor to resize cell parameters.
    path:
        Path to the ref_XYZ.extxyz file.
    x_min: int
        Minimum value for the plot x-axis.
    x_max: int
        Maximum value for the plot x-axis.
    bin_width: float
        Width of the plot bins.

    """
    plt.xlabel("Forces")
    plt.ylabel("Count")
    bins = np.arange(x_min, x_max + bin_width, bin_width)
    plot_total = []

    # TODO split data collection and plotting

    plot_data = []
    with open(path + "/ref_" + str(cell_factor).replace(".", "") + ".extxyz") as file:
        for line in file:
            # Split the line into columns
            columns = line.split()

            # Check if the line has exactly 10 columns
            if len(columns) == 10:
                # Extract the last three columns
                data = columns[-3:]
                norm_data = np.linalg.norm(data, axis=-1)
                plot_data.append(norm_data)

        plt.hist(plot_data, bins=bins, edgecolor="black")
        plt.title(f"Data for factor {cell_factor}")

        plt.savefig("data_factor_" + str(cell_factor).replace(".", "") + ".png")

        plot_total += plot_data
    plt.hist(plot_total, bins=bins, edgecolor="black")
    plt.title("Data")

    plt.savefig("total_data.png")


@job
def get_supercell_job(structure: Structure, supercell_matrix: Matrix3D):
    """
    Create a job to get the supercell.

    Parameters
    ----------
    structure: Structure
        The pymatgen structure object.
    supercell_matrix: Matrix3D
        The matrix to generate the supercell.

    Returns
    -------
    supercell: Structure
        A pymatgen structure object.

    """
    supercell = get_supercell(
        unitcell=get_phonopy_structure(structure), supercell_matrix=supercell_matrix
    )
    return get_pmg_structure(supercell)


@job(data=[Structure])
def generate_randomized_structures(
    structure: Structure,
    supercell_matrix: Matrix3D | None = None,
    distort_type: int = 0,
    n_structures: int = 10,
    volume_scale_factor_range: list[float] | None = None,
    volume_custom_scale_factors: list[float] | None = None,
    min_distance: float = 1.5,
    angle_percentage_scale: float = 10,
    angle_max_attempts: int = 1000,
    rattle_type: int = 0,
    rattle_std: float = 0.01,
    rattle_seed: int = 42,
    rattle_mc_n_iter: int = 10,
    w_angle: list[float] | None = None,
):
    """
    Take in a pymatgen Structure object and generates angle/volume distorted + rattled structures.

    Parameters
    ----------
    structure: Structure.
        Pymatgen structures object.
    supercell_matrix: Matrix3D.
        Matrix for obtaining the supercell.
    distort_type: int.
        0- volume distortion, 1- angle distortion, 2- volume and angle distortion. Default=0.
    n_structures: int.
        Total number of distorted structures to be generated.
        Must be provided if distorting volume without specifying a range, or if distorting angles.
        Default=10.
    volume_scale_factor_range: list[float]
        [min, max] of volume scale factors.
        e.g. [0.90, 1.10] will distort volume +-10%.
    volume_custom_scale_factors: list[float]
        Specify explicit scale factors (if range is not specified).
        If None, will default to [0.90, 0.95, 0.98, 0.99, 1.01, 1.02, 1.05, 1.10].
    min_distance: float
        Minimum separation allowed between any two atoms.
        Default= 1.5A.
    angle_percentage_scale: float
        Angle scaling factor.
        Default= 10 will randomly distort angles by +-10% of original value.
    angle_max_attempts: int.
        Maximum number of attempts to distort structure before aborting.
        Default=1000.
    w_angle: list[float]
        List of angle indices to be changed i.e. 0=alpha, 1=beta, 2=gamma.
        Default= [0, 1, 2].
    rattle_type: int.
        0- standard rattling, 1- Monte-Carlo rattling. Default=0.
    rattle_std: float.
        Rattle amplitude (standard deviation in normal distribution).
        Default=0.01.
        Note that for MC rattling, displacements generated will roughly be
        rattle_mc_n_iter**0.5 * rattle_std for small values of n_iter.
    rattle_seed: int.
        Seed for setting up NumPy random state from which random numbers are generated.
        Default=42.
    rattle_mc_n_iter: int.
        Number of Monte Carlo iterations.
        Larger number of iterations will generate larger displacements.
        Default=10.

    Returns
    -------
    Response.output.
        Volume or angle-distorted structures with rattled atoms.
    """
    if supercell_matrix is None:
        supercell_matrix = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    supercell = get_supercell(
        unitcell=get_phonopy_structure(structure),
        supercell_matrix=supercell_matrix,
    )
    structure = get_pmg_structure(supercell)

    # distort cells by volume or angle
    if distort_type == 0:
        distorted_cells = scale_cell(
            structure=structure,
            volume_scale_factor_range=volume_scale_factor_range,
            n_structures=n_structures,
            volume_custom_scale_factors=volume_custom_scale_factors,
        )
    elif distort_type == 1:
        distorted_cells = random_vary_angle(
            structure=structure,
            min_distance=min_distance,
            angle_percentage_scale=angle_percentage_scale,
            w_angle=w_angle,
            n_structures=n_structures,
            angle_max_attempts=angle_max_attempts,
        )
    elif distort_type == 2:
        initial_distorted_cells = scale_cell(
            structure=structure,
            volume_scale_factor_range=volume_scale_factor_range,
            n_structures=n_structures,
            volume_custom_scale_factors=volume_custom_scale_factors,
        )
        distorted_cells = []
        for cell in initial_distorted_cells:
            distorted_cell = random_vary_angle(
                structure=cell,
                min_distance=min_distance,
                angle_percentage_scale=angle_percentage_scale,
                w_angle=w_angle,
                n_structures=1,
                angle_max_attempts=angle_max_attempts,
            )
            distorted_cells.append(distorted_cell)
        distorted_cells = list(chain.from_iterable(distorted_cells))
    else:
        raise TypeError("distort_type is not recognised")

    # distorted_cells=list(chain.from_iterable(distorted_cells))

    # rattle cells by standard or mc
    rattled_cells = (
        [
            std_rattle(
                structure=cell,
                n_structures=1,
                rattle_std=rattle_std,
                rattle_seed=rattle_seed + icell,
            )
            for icell, cell in enumerate(distorted_cells)
        ]
        if rattle_type == 0
        else (
            [
                mc_rattle(
                    structure=cell,
                    n_structures=1,
                    rattle_std=rattle_std,
                    min_distance=min_distance,
                    rattle_seed=rattle_seed + icell,
                    rattle_mc_n_iter=rattle_mc_n_iter,
                )
                for icell, cell in enumerate(distorted_cells)
            ]
            if rattle_type == 1
            else None
        )
    )

    if rattled_cells is None:
        raise TypeError("rattle_type is not recognized")

    return list(chain.from_iterable(rattled_cells))


@job
def sample_data(
    selection_method: Literal[
        "cur", "bcur1s", "bcur2i", "random", "uniform"
    ] = "random",
    num_of_selection: int = 5,
    bcur_params: dict | None = None,
    dir: list[str] | str | None = None,
    structure: list[Structure] | list[list[Structure]] | None = None,
    traj_path: list | None = None,
    isolated_atom_energies: dict | None = None,
    random_seed: int = None,
    remove_traj_files: bool = False,
) -> list[Structure]:
    """
    Job to sample training configurations from trajectories of MD/RSS.

    Parameters
    ----------
    selection_method : Literal['cur', 'bcur1s', 'bcur2s', 'random', 'uniform']
       Method for selecting samples. Options include:

        - 'cur': Pure CUR selection.

        - 'bcur': Boltzmann flat histogram in enthalpy, then CUR.

            - 'bcur1s': Execute bcur with one shot (1s)
            - 'bcur2i': Execute bcur with two iterations (2i)

        - 'random': Random selection.
        - 'uniform': Uniform selection.

    num_of_selection: int
       Number of structures to be sampled.
    bcur_params: dict
        Parameters for Boltzmann CUR selection. The default dictionary includes:

        - 'soap_paras': SOAP descriptor parameters

            - 'l_max': int, Maximum degree of spherical harmonics (default 12).
            - 'n_max': int, Maximum number of radial basis functions (default 12).
            - 'atom_sigma': float, Width of Gaussian smearing (default 0.0875).
            - 'cutoff': float, Radial cutoff distance (default 10.5).
            - 'cutoff_transition_width': float, Width of the transition region (default 1.0).
            - 'zeta': float, Exponent for dot-product SOAP kernel (default 4.0).
            - 'average': bool, Whether to average the SOAP vectors (default True).
            - 'species': bool, Whether to consider species information (default True).

        - 'kt': float, Temperature in eV for Boltzmann weighting (default 0.3).
        - 'frac_of_bcur': float, Fraction of Boltzmann CUR selections (default 0.8).
        - 'bolt_max_num': int, Maximum number of Boltzmann selections (default 3000).
        - 'kernel_exp': float, Exponent for the kernel (default 4.0).
        - 'energy_label': str, Label for the energy data (default 'energy').

    dir: str
        Directory containing trajectory files for MD/RSS simulations. Default is None.
    structure: list[Structure]
        List of structures for sampling. Default is None.
    traj_path: list[list[str]]
        List of lists containing trajectory paths. Default is None.
    isolated_atom_energies: dict
        Dictionary of isolated energy values for species. Required for 'boltzhist_cur'
        selection method. Default is None.
    random_seed: int, optional
        Seed for random number generation, ensuring reproducibility of sampling.
    remove_traj_files: bool
        Remove all trajectory files raised by RSS to save memory

    Returns
    -------
    list of ase.Atoms
        The selected atoms.
    """
    default_bcur_params = {
        "soap_paras": {
            "l_max": 12,
            "n_max": 12,
            "atom_sigma": 0.0875,
            "cutoff": 10.5,
            "cutoff_transition_width": 1.0,
            "zeta": 4.0,
            "average": True,
            "species": True,
        },
        "kt": 0.3,
        "frac_of_bcur": 0.8,
        "bolt_max_num": 3000,
        "kernel_exp": 4.0,
        "energy_label": "energy",
    }

    if bcur_params is not None:
        default_bcur_params.update(bcur_params)

    bcur_params = default_bcur_params
    pressures = None

    if dir is not None:
        if isinstance(dir, list):
            atoms = [read(i, index=":") for i in dir]
            atoms = flatten(atoms, recursive=True)
        else:
            atoms = read(dir, index=":")

    elif structure is not None:
        if isinstance(structure[0], list):
            structure = flatten(structure, recursive=False)
        atoms = [AseAtomsAdaptor().get_atoms(at) for at in structure]

    else:
        atoms, pressures = handle_rss_trajectory(traj_path, remove_traj_files)

    if selection_method in {"cur", "bcur1s", "bcur2i"}:
        n_species = ElementCollection(
            flatten(atoms, recursive=True)
        ).get_number_of_species()
        species_Z = ElementCollection(flatten(atoms, recursive=True)).get_species_Z()

        if not isinstance(bcur_params["soap_paras"], dict):
            raise TypeError("soap_paras must be a dictionary")
        if not isinstance(bcur_params["kt"], float):
            raise TypeError("kt must be a float")
        if not isinstance(bcur_params["frac_of_bcur"], float):
            raise TypeError("frac_of_bcur must be a float")
        if not isinstance(bcur_params["bolt_max_num"], int):
            raise TypeError("bolt_max_num must be an integer")
        if not isinstance(bcur_params["kernel_exp"], float):
            raise TypeError("kernel_exp must be a float")
        if not isinstance(bcur_params["energy_label"], str):
            raise TypeError("energy_label must be a string")

        soap_paras = bcur_params["soap_paras"]
        descriptor = create_soap_descriptor(soap_paras, n_species, species_Z)

        if selection_method == "cur":
            selected_atoms = cur_select(
                atoms=atoms,
                selected_descriptor=descriptor,
                kernel_exp=bcur_params["kernel_exp"],
                select_nums=num_of_selection,
                stochastic=True,
                random_seed=random_seed,
            )

        elif selection_method in {"bcur1s", "bcur2i"}:
            if isolated_atom_energies is not None:
                isolated_atom_energies = {
                    int(k): v for k, v in isolated_atom_energies.items()
                }
            else:
                raise ValueError("Please provide the energy of isolated atoms!")

            if selection_method == "bcur1s":
                selected_atoms = boltzhist_cur_one_shot(
                    atoms=atoms,
                    isolated_atom_energies=isolated_atom_energies,
                    bolt_frac=bcur_params["frac_of_bcur"],
                    bolt_max_num=bcur_params["bolt_max_num"],
                    cur_num=num_of_selection,
                    kernel_exp=bcur_params["kernel_exp"],
                    kt=bcur_params["kt"],
                    energy_label=bcur_params["energy_label"],
                    pressures=pressures,
                    descriptor=descriptor,
                    random_seed=random_seed,
                )
            else:
                selected_atoms = boltzhist_cur_dual_iter(
                    atoms=atoms,
                    isolated_atom_energies=isolated_atom_energies,
                    bolt_frac=bcur_params["frac_of_bcur"],
                    bolt_max_num=bcur_params["bolt_max_num"],
                    cur_num=num_of_selection,
                    kernel_exp=bcur_params["kernel_exp"],
                    kt=bcur_params["kt"],
                    energy_label=bcur_params["energy_label"],
                    pressures=pressures,
                    descriptor=descriptor,
                    random_seed=random_seed,
                )

        if selected_atoms is None:
            raise ValueError(
                "Unable to sample correctly. Please recheck the parameters!"
            )

        selected_atoms = [AseAtomsAdaptor().get_structure(at) for at in selected_atoms]

    elif selection_method == "random":
        if random_seed is not None:
            np.random.seed(random_seed)

        structure = [AseAtomsAdaptor().get_structure(at) for at in atoms]

        try:
            selection = np.random.choice(
                len(structure), num_of_selection, replace=False
            )
            selected_atoms = [at for i, at in enumerate(structure) if i in selection]

        except ValueError:
            logging.error(
                "The number of selected structures must be less than the total!"
            )
            traceback.print_exc()

    elif selection_method == "uniform":
        try:
            indices = np.linspace(0, len(atoms) - 1, num_of_selection, dtype=int)
            structure = [AseAtomsAdaptor().get_structure(at) for at in atoms]
            selected_atoms = [structure[idx] for idx in indices]

        except ValueError:
            logging.error(
                "The number of selected structures must be less than the total!"
            )
            traceback.print_exc()

    if selected_atoms is None:
        raise ValueError("Unable to sample correctly. Please recheck the parameters!")

    return selected_atoms


@job
def collect_dft_data(
    dft_ref_file: str = "dft_ref.extxyz",
    rss_group: str = "RSS",
    dft_dirs: dict | None = None,
) -> dict:
    """
    Collect VASP data from specified directories.

    Parameters
    ----------
    dft_ref_file : str
        Reference file for DFT-labelled data. Default is 'dft_ref.extxyz'.
    rss_group : str
        Group name for GAP RSS. Default is 'RSS'.
    dft_dirs : dict
        Dictionary containing DFT calculation directories and configuration types. Should have keys:

        - 'dirs_of_dft': list
            List of directories containing DFT data.
        - 'config_type': list
            List of configuration types corresponding to each directory.

    Returns
    -------
    dict:
        A dictionary containing

        - 'dft_ref_dir': Directory of the dft reference file.
        - 'isolated_atom_energies': Isolated energy values.
    """
    if dft_dirs is None:
        raise ValueError(
            "dft_dirs must be provided and should contain 'dirs_of_dft' and 'config_type' keys."
        )

    if "dirs_of_dft" not in dft_dirs or "config_type" not in dft_dirs:
        raise ValueError("dft_dirs must contain 'dirs_of_dft' and 'config_type' keys.")

    dirs = [safe_strip_hostname(value) for value in dft_dirs["dirs_of_dft"]]
    config_types = dft_dirs["config_type"]

    logging.info("Attempting collecting DFT...")

    if dirs is None:
        raise ValueError("dft_dir must be specified if collect_dft is True")

    atoms = []
    isolated_atom_energies = {}

    for i, val in enumerate(dirs):
        at = None
        # TODO: think about fall back when we use ml potential outputs instead
        has_vasp_output = os.path.exists(os.path.join(val, "vasprun.xml.gz"))
        has_castep_output = os.path.exists(os.path.join(val, "castep.castep.gz"))
        has_ase_output = os.path.exists(os.path.join(val, "final_atoms_object.xyz"))

        if has_vasp_output or has_castep_output or has_ase_output:
            if has_vasp_output:
                converged = check_convergence_vasp(os.path.join(val, "vasprun.xml.gz"))
                if converged:
                    at = read(os.path.join(val, "vasprun.xml.gz"), index=":")
                else:
                    logging.warning(
                        f"Calculation did not converge for path: {os.path.join(val, 'vasprun.xml.gz')}"
                    )
            elif has_castep_output:
                converged = check_convergence_castep(
                    os.path.join(val, "castep.castep.gz")
                )
                if converged:
                    at = read(os.path.join(val, "castep.castep.gz"), index=":")
                else:
                    logging.warning(
                        f"Calculation did not converge for path: {os.path.join(val, 'castep.castep.gz')}"
                    )
            elif has_ase_output:
                try:
                    logging.info("read ase")
                    at = read(
                        os.path.join(val, "final_atoms_object.xyz"),
                        index=":",
                        format="extxyz",
                    )
                except AssertionError:
                    pass
            if at is not None:
                for at_i in at:

                    logging.warning(at_i.get_stress())

                    virial_list = (
                        -voigt_6_to_full_3x3_stress(at_i.get_stress())
                        * at_i.get_volume()
                    )
                    at_i.info["REF_virial"] = " ".join(map(str, virial_list.flatten()))
                    del at_i.calc.results["stress"]
                    at_i.arrays["REF_forces"] = at_i.calc.results["forces"]
                    del at_i.calc.results["forces"]
                    at_i.info["REF_energy"] = at_i.calc.results["free_energy"]
                    del at_i.calc.results["energy"]
                    del at_i.calc.results["free_energy"]
                    atoms.append(at_i)
                    at_i.info["config_type"] = config_types[i]
                    if (
                        at_i.info["config_type"] != "dimer"
                        and at_i.info["config_type"] != "IsolatedAtom"
                    ):
                        at_i.pbc = True
                        at_i.info["rss_group"] = rss_group
                    else:
                        at_i.info["rss_nonperiodic"] = "T"

                    if at_i.info["config_type"] == "IsolatedAtom":
                        at_ids = at_i.get_atomic_numbers()
                        # array_key = at_ids.tostring()
                        isolated_atom_energies[int(at_ids[0])] = at_i.info["REF_energy"]

    logging.info(f"Total {len(atoms)} structures from VASP are exactly collected.")

    write(dft_ref_file, atoms, format="extxyz", parallel=False)

    dir_path = Path.cwd()

    dft_ref_dir = os.path.join(dir_path, dft_ref_file)

    return {
        "dft_ref_dir": dft_ref_dir,
        "isolated_atom_energies": isolated_atom_energies,
    }


def check_convergence_vasp(path: str) -> bool:
    """
    Check if VASP calculation has converged.

    Parameters
    ----------
    path: str
        Path to the vasp output file to check convergence.

    Return
    ------
    bool
        True if a run is converged both ionically and electronically.
    """
    vasprun = Vasprun(path)
    converged_e = vasprun.converged_electronic
    converged_i = vasprun.converged_ionic

    return converged_e and converged_i


def check_convergence_castep(castep_gz: str) -> bool:
    """
    Check if CASTEP calculation has converged.

    Parameters
    ----------
    castep_gz : str
        Path to the .castep.gz file.

    Returns
    -------
    bool
        True if converged, False otherwise.
    """
    with gzip.open(castep_gz, "rt") as f:
        for line in f:
            if "electronic minimisation did not converge" in line:
                return False

    return True


def safe_strip_hostname(value):
    """
    Strip the hostname from a given path or URL.

    Parameters
    ----------
    value: str
        The path or URL from which to strip the hostname.

    Returns
    -------
    Optional[str]
        The path or URL without the hostname if the operation is successful,
        otherwise None.
    """
    try:
        return strip_hostname(value)
    except Exception as e:
        print(f"Error processing '{value}': {e}")
        return None


@job
def preprocess_data(
    dft_ref_dir: str,
    test_ratio: float | None = None,
    regularization: bool = False,
    retain_existing_sigma: bool = False,
    scheme: str = "linear-hull",
    element_order: list | None = None,
    distillation: bool = False,
    force_max: float = 40,
    force_label: str = "REF_forces",
    energy_label: str = "REF_energy",
    pre_database_dir: str | None = None,
    reg_minmax: list[tuple] | None = None,
    isolated_atom_energies: dict | None = None,
) -> Path:
    """
    Preprocesse data to before fiting machine learning models.

    This function handles tasks such as splitting the dataset,
    applying regularization, accumulating database, and filtering
    structures based on maximum force values.

    Parameters
    ----------
    dft_ref_dir: str
        Path to the directory containing the reference DFT calculation data.
    test_ratio: float
        The proportion of the test set after splitting the data.
        If None, no splitting will be performed.
    regularization: bool
        If true, apply regularization. This only works for GAP.
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.
    scheme: str
        Scheme to use for regularization.
    element_order:
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2).
        This value is useful when constructing high-dimensional convex hulls based on the
        "volume-stoichiometry" scheme. Specially, if the dataset contains compounds with
        different numbers of constituent elements (e.g., both binary and ternary structures),
        this value must be explicitly set to ensure the convex hull is constructed consistently.
    distillation: bool
        If True, apply data distillation.
    force_max: float
        Maximum force value to exclude structures.
    force_label: str
        The label of force values to use for distillation.
    energy_label: str
        The label of energy values to use for distillation.
    pre_database_dir : str
        Directory where the previous database was saved.
    reg_minmax: list[tuple]
        A list of tuples representing the minimum and maximum
        values for regularization.
    isolated_atom_energies: dict
        A dictionary containing isolated energy values for different species.

    Returns
    -------
    Path
        The current working directory.
    """
    atoms = (
        data_distillation(dft_ref_dir, force_max, force_label)
        if distillation
        else read(dft_ref_dir, index=":")
    )

    if test_ratio == 0 or test_ratio is None:
        train_structures, test_structures = atoms, atoms
    else:
        train_structures, test_structures = stratified_dataset_split(
            atoms, test_ratio, energy_label
        )

    if pre_database_dir and os.path.exists(pre_database_dir):
        files_to_copy = ["train.extxyz", "test.extxyz"]
        current_working_directory = os.getcwd()

        for file_name in files_to_copy:
            source_file_path = os.path.join(pre_database_dir, file_name)
            destination_file_path = os.path.join(current_working_directory, file_name)
            if os.path.exists(source_file_path):
                shutil.copy(source_file_path, destination_file_path)
                print(f"File {file_name} has been copied to {destination_file_path}")

    write("train.extxyz", train_structures, format="extxyz", append=True)
    write("test.extxyz", test_structures, format="extxyz", append=True)

    if regularization:
        atoms_reg: list[Atoms] = read("train.extxyz", index=":")

        if reg_minmax is None:
            reg_minmax = [(0.1, 1), (0.001, 0.1), (0.0316, 0.316), (0.0632, 0.632)]

        atom_with_sigma = set_custom_sigma(
            atoms=atoms_reg,
            reg_minmax=reg_minmax,
            isolated_atom_energies=isolated_atom_energies,
            scheme=scheme,
            element_order=element_order,
            retain_existing_sigma=retain_existing_sigma,
        )

        write("train.extxyz", atom_with_sigma, format="extxyz")

    return Path.cwd()



"""Flows to create training data for ML potentials."""

from dataclasses import dataclass, field

from atomate2.common.jobs.phonons import run_phonon_displacements
from atomate2.forcefields.flows.phonons import PhononMaker as FFPhononMaker
from atomate2.forcefields.jobs import (
    ForceFieldRelaxMaker,
    ForceFieldStaticMaker,
)
from atomate2.vasp.flows.core import DoubleRelaxMaker
from atomate2.vasp.flows.phonons import PhononMaker
from atomate2.vasp.jobs.base import BaseVaspMaker
from atomate2.vasp.jobs.core import StaticMaker, TightRelaxMaker
from atomate2.vasp.jobs.phonons import PhononDisplacementMaker
from atomate2.vasp.sets.base import VaspInputGenerator
from atomate2.vasp.sets.core import StaticSetGenerator, TightRelaxSetGenerator
from jobflow import Flow, Maker, Response, job
from pymatgen.core import Molecule, Site
from pymatgen.core.structure import Species, Structure

from autoplex.data.common.jobs import generate_randomized_structures
from autoplex.data.phonons.jobs import reduce_supercell_size_job
from autoplex.data.phonons.utils import (
    ml_phonon_maker_preparation,
    reduce_supercell_size,
)

__all__ = [
    "DFTPhononMaker",
    "IsoAtomMaker",
    "IsoAtomStaticMaker",
    "MLPhononMaker",
    "RandomStructuresDataGenerator",
    "TightDFTStaticMaker",
]


@dataclass
class TightDFTStaticMaker(PhononDisplacementMaker):
    """Adapted phonon displacement maker for static calculation.

    The input set used is same as PhononDisplacementMaker.
    Only difference is Spin polarization is switched off and Gaussian smearing is used

    Parameters
    ----------
    name : str
        The job name.
    input_set_generator : .VaspInputGenerator
        Generator used to make the input set.
    write_input_set_kwargs : dict
        Keyword arguments that will get passed to :obj:`.write_vasp_input_set`.
    copy_vasp_kwargs : dict
        Keyword arguments that will get passed to :obj:`.copy_vasp_outputs`.
    run_vasp_kwargs : dict
        Keyword arguments that will get passed to :obj:`.run_vasp`.
    task_document_kwargs : dict
        Keyword arguments that will get passed to :obj:`.TaskDoc.from_directory`.
    stop_children_kwargs : dict
        Keyword arguments that will get passed to :obj:`.should_stop_children`.
    write_additional_data : dict
        Additional data to write to the current directory. Given as a dict of
        {filename: data}. Note that if using FireWorks, dictionary keys cannot contain
        the "." character which is typically used to denote file extensions. To avoid
        this, use the ":" character, which will automatically be converted to ".". E.g.
        ``{"my_file:txt": "contents of the file"}``.
    """

    name: str = "dft static"
    run_vasp_kwargs: dict = field(default_factory=lambda: {"handlers": ()})

    input_set_generator: VaspInputGenerator = field(
        default_factory=lambda: StaticSetGenerator(
            user_incar_settings={
                "ALGO": "Normal",  # not switching to Fast because it's not precise enough for the fit
                "IBRION": -1,
                "ISPIN": 1,
                "ISMEAR": 0,
                "ISIF": 3,
                "ENCUT": 700,
                "EDIFF": 1e-7,
                "LAECHG": False,
                "LREAL": False,
                "NSW": 0,
                "LCHARG": False,  # Do not write the CHGCAR file
                "LWAVE": False,  # Do not write the WAVECAR file
                "LVTOT": False,  # Do not write LOCPOT file
                "LORBIT": None,  # remove Lorbit
                "LOPTICS": False,  # No PCDAT file
                "SIGMA": 0.05,
                "ISYM": 0,
                "SYMPREC": 1e-9,
                "KSPACING": 0.2,
                # To be removed
                "NPAR": 4,
            },
            auto_ispin=False,
        )
    )


@dataclass
class DFTPhononMaker(PhononMaker):
    """
    Adapted PhononMaker to calculate harmonic phonons with VASP and Phonopy.

    The input set used is same as PhononMaker from atomate2.
    Only difference is Spin polarization is switched off and Gaussian smearing is used

    Parameters
    ----------
    name : str = "phonon"
        Name of the flows produced by this maker.
    sym_reduce : bool = True
        Whether to reduce the number of deformations using symmetry.
    symprec : float = 1e-4
        Symmetry precision to use in the
        reduction of symmetry to find the primitive/conventional cell
        (use_primitive_standard_structure, use_conventional_standard_structure)
        and to handle all symmetry-related tasks in phonopy
    displacement: float = 0.01
        Displacement distance for phonons
    min_length: float = 20.0
        Minimum length of the supercell that will be built
    prefer_90_degrees: bool = True
        If set to True, supercell algorithm will first try to find a supercell
        with 3 90 degree angles
    get_supercell_size_kwargs: dict = {}
        Keyword arguments that will be passed to get_supercell_size to determine supercell size
    use_symmetrized_structure: str or None = None
        Allowed strings: "primitive", "conventional", None

        - "primitive" will enforce to start the phonon computation
          from the primitive standard structure
          according to Setyawan, W., & Curtarolo, S. (2010).
          High-throughput electronic band structure calculations:
          Challenges and tools. Computational Materials Science,
          49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010.
          This makes it possible to use certain k-path definitions
          with this workflow. Otherwise, we must rely on seekpath
        - "conventional" will enforce to start the phonon computation
          from the conventional standard structure
          according to Setyawan, W., & Curtarolo, S. (2010).
          High-throughput electronic band structure calculations:
          Challenges and tools. Computational Materials Science,
          49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010.
          We will however use seekpath and primitive structures
          as determined by from phonopy to compute the phonon band structure
    bulk_relax_maker : .BaseVaspMaker or None
        Maker to perform a tight relaxation on the bulk.
        Set to ``None`` to skip the
        bulk relaxation
    static_energy_maker : .BaseVaspMaker or None
        Maker to perform the computation of the DFT energy on the bulk.
        Set to ``None`` to skip the
        static energy computation
    born_maker: .BaseVaspMaker or None
        Maker to compute the BORN charges.
    phonon_displacement_maker : .BaseVaspMaker or None
        Maker used to compute the forces for a supercell.
    generate_frequencies_eigenvectors_kwargs : dict
        Keyword arguments passed to :obj:`generate_frequencies_eigenvectors`.
    create_thermal_displacements: bool
        Arg that determines if thermal_displacement_matrices are computed
    kpath_scheme: str = "seekpath"
        Scheme to generate kpoints. Please be aware that
        you can only use seekpath with any kind of cell
        Otherwise, please use the standard primitive structure
        Available schemes are:
        "seekpath", "hinuma", "setyawan_curtarolo", "latimer_munro".
        "seekpath" and "hinuma" are the same definition but
        seekpath can be used with any kind of unit cell as
        it relies on phonopy to handle the relationship
        to the primitive cell and not pymatgen
    code: str = "vasp"
        Determines the DFT code. currently only vasp is implemented.
        This keyword might enable the implementation of other codes
        in the future
    store_force_constants: bool
        If True, force constants will be stored
    """

    name: str = "dft phonon"
    sym_reduce: bool = True
    symprec: float = 1e-4
    displacement: float = 0.01
    min_length: float | None = 20.0
    max_length: float | None = 30.0
    prefer_90_degrees: bool = True
    allow_orthorhombic: bool = False
    get_supercell_size_kwargs: dict = field(
        default_factory=lambda: {"max_atoms": 800, "step_size": 1.0}
    )
    use_symmetrized_structure: str | None = None
    create_thermal_displacements: bool = False
    store_force_constants: bool = False
    generate_frequencies_eigenvectors_kwargs: dict = field(
        default_factory=lambda: {"tol_imaginary_modes": 1e-1}
    )
    bulk_relax_maker: BaseVaspMaker | None = field(
        default_factory=lambda: DoubleRelaxMaker.from_relax_maker(
            TightRelaxMaker(
                run_vasp_kwargs={"handlers": {}},
                input_set_generator=TightRelaxSetGenerator(
                    user_incar_settings={
                        "ALGO": "Normal",
                        "ISPIN": 1,
                        "LAECHG": False,
                        "ISMEAR": 0,
                        "ENCUT": 700,
                        "ISYM": 0,
                        "SIGMA": 0.05,
                        "LCHARG": False,  # Do not write the CHGCAR file
                        "LWAVE": False,  # Do not write the WAVECAR file
                        "LVTOT": False,  # Do not write LOCPOT file
                        "LORBIT": None,  # No output of projected or partial DOS in EIGENVAL, PROCAR and DOSCAR
                        "LOPTICS": False,  # No PCDAT file
                        # to be removed
                        "NPAR": 4,
                    }
                ),
            )
        ),
    )
    static_energy_maker: BaseVaspMaker | None = None

    phonon_displacement_maker: BaseVaspMaker | None = field(
        default_factory=TightDFTStaticMaker
    )


@dataclass
class RandomStructuresDataGenerator(Maker):
    """
    Maker to generate DFT labelled training data for ML potential fitting based on random atomic displacements.

    This Maker performs the two following steps:
    1. Generates supercells from the provided structure and randomly displaces the atomic positions using ase rattle.
    (randomized unit cells can be generated additionally).
    2. Performs the static DFT (VASP) calculations on the randomized cells.

    Parameters
    ----------
    name : str
        Name of the flows produced by this maker.
    displacement_maker: .BaseVaspMaker or None
        Maker used for a static calculation for a supercell.
    code: str
        Determines the dft code. currently only vasp is implemented.
        This keyword might enable the implementation of other codes
        in the future
    n_structures : int.
        Total number of distorted structures to be generated.
        Must be provided if distorting volume without specifying a range, or if distorting angles.
        Default=10.
    uc: bool.
        If True, will use the unit cells of initial randomly displaced
        structures and add phonon static computation jobs to the flow
    distort_type : int.
        0- volume distortion, 1- angle distortion, 2- volume and angle distortion. Default=0.
    min_distance: float
        Minimum separation allowed between any two atoms.
        Default= 1.5A.
    angle_percentage_scale: float
        Angle scaling factor.
        Default= 10 will randomly distort angles by +-10% of original value.
    angle_max_attempts: int.
        Maximum number of attempts to distort structure before aborting.
        Default=1000.
    w_angle: list[float]
        List of angle indices to be changed i.e. 0=alpha, 1=beta, 2=gamma.
        Default= [0, 1, 2].
    rattle_type: int.
        0- standard rattling, 1- Monte-Carlo rattling. Default=0.
    rattle_std: float.
        Rattle amplitude (standard deviation in normal distribution).
        Default=0.01.
        Note that for MC rattling, displacements generated will roughly be
        rattle_mc_n_iter**0.5 * rattle_std for small values of n_iter.
    rattle_seed: int.
        Seed for setting up NumPy random state from which random numbers are generated.
        Default=42.
    rattle_mc_n_iter: int.
        Number of Monte Carlo iterations.
        Larger number of iterations will generate larger displacements.
        Default=10.
    supercell_settings: dict
        Settings for supercells.
    """

    name: str = "RandomStruturesDataGeneratorForML"
    displacement_maker: BaseVaspMaker | None = field(
        default_factory=TightDFTStaticMaker
    )
    bulk_relax_maker: BaseVaspMaker = field(
        default_factory=lambda: TightRelaxMaker(
            run_vasp_kwargs={"handlers": {}},
            input_set_generator=TightRelaxSetGenerator(
                user_incar_settings={
                    "ALGO": "Normal",
                    "ISPIN": 1,
                    "LAECHG": False,
                    "ISYM": 0,  # to be changed
                    "ISMEAR": 0,
                    "SIGMA": 0.05,  # to be changed back
                    "LCHARG": False,  # Do not write the CHGCAR file
                    "LWAVE": False,  # Do not write the WAVECAR file
                    "LVTOT": False,  # Do not write LOCPOT file
                    "LORBIT": None,  # No output of projected or partial DOS in EIGENVAL, PROCAR and DOSCAR
                    "LOPTICS": False,  # No PCDAT file
                    # to be removed
                    "NPAR": 4,
                }
            ),
        )
    )
    code: str = "vasp"
    uc: bool = False
    distort_type: int = 0
    n_structures: int = 10
    min_distance: float = 1.5
    angle_percentage_scale: float = 10
    angle_max_attempts: int = 1000
    rattle_type: int = 0
    rattle_std: float = 0.01
    rattle_seed: int = 42
    rattle_mc_n_iter: int = 10
    w_angle: list[float] | None = None
    supercell_settings: dict | None = field(
        default_factory=lambda: {"min_length": 15, "max_length": 20}
    )

    def make(
        self,
        structure: Structure,
        mp_id: str,
        volume_custom_scale_factors: list[float] | None = None,
        volume_scale_factor_range: list[float] | None = None,
    ):
        """
        Make a flow to generate rattled structures reference DFT data.

        Parameters
        ----------
        structure :
            The pymatgen structures drawn from the Materials Project.
        mp_id: str
            Materials Project IDs
        volume_scale_factor_range : list[float]
            [min, max] of volume scale factors.
            e.g. [0.90, 1.10] will distort volume +-10%.
        volume_custom_scale_factors : list[float]
            Specify explicit scale factors (if range is not specified).
            If None, will default to [0.90, 0.95, 0.98, 0.99, 1.01, 1.02, 1.05, 1.10].
        """
        if self.supercell_settings is None:
            self.supercell_settings = field(
                default_factory=lambda: {"min_length": 15, "max_length": 20}
            )
        jobs = []  # initializing empty job list
        outputs = []
        if self.bulk_relax_maker is not None:
            relaxed = self.bulk_relax_maker.make(structure)
            jobs.append(relaxed)
            structure = relaxed.output.structure

        supercell_matrix = self.supercell_settings.get(mp_id, {}).get(
            "supercell_matrix"
        )
        if not supercell_matrix:
            supercell_matrix_job = reduce_supercell_size_job(
                structure=structure,
                min_length=self.supercell_settings.get("min_length", 12),
                max_length=self.supercell_settings.get("max_length", 20),
                fallback_min_length=self.supercell_settings.get(
                    "fallback_min_length", 10
                ),
                max_atoms=self.supercell_settings.get("max_atoms", 500),
                min_atoms=self.supercell_settings.get("min_atoms", 50),
                step_size=self.supercell_settings.get("step_size", 1.0),
            )
            jobs.append(supercell_matrix_job)
            supercell_matrix = supercell_matrix_job.output

        random_rattle_sc = generate_randomized_structures(
            structure=structure,
            supercell_matrix=supercell_matrix,
            distort_type=self.distort_type,
            n_structures=self.n_structures,
            volume_custom_scale_factors=volume_custom_scale_factors,
            volume_scale_factor_range=volume_scale_factor_range,
            rattle_std=self.rattle_std,
            min_distance=self.min_distance,
            angle_percentage_scale=self.angle_percentage_scale,
            angle_max_attempts=self.angle_max_attempts,
            rattle_type=self.rattle_type,
            rattle_seed=self.rattle_seed,
            rattle_mc_n_iter=self.rattle_mc_n_iter,
            w_angle=self.w_angle,
        )
        jobs.append(random_rattle_sc)
        # perform the phonon displaced calculations for randomized displaced structures.
        #  The original structure is only needed to keep track of initial structure.
        vasp_random_sc_displacement_calcs = run_phonon_displacements(
            displacements=random_rattle_sc.output,  # pylint: disable=E1101
            structure=structure,
            supercell_matrix=None,
            phonon_maker=self.displacement_maker,
        )

        jobs.append(vasp_random_sc_displacement_calcs)
        outputs.append(vasp_random_sc_displacement_calcs.output["dirs"])

        if self.uc is True:
            random_rattle = generate_randomized_structures(
                structure=structure,
                supercell_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
                distort_type=self.distort_type,
                n_structures=self.n_structures,
                volume_custom_scale_factors=volume_custom_scale_factors,
                volume_scale_factor_range=volume_scale_factor_range,
                rattle_std=self.rattle_std,
                min_distance=self.min_distance,
                angle_percentage_scale=self.angle_percentage_scale,
                angle_max_attempts=self.angle_max_attempts,
                rattle_type=self.rattle_type,
                rattle_seed=self.rattle_seed,
                rattle_mc_n_iter=self.rattle_mc_n_iter,
                w_angle=self.w_angle,
            )
            jobs.append(random_rattle)
            vasp_random_displacement_calcs = run_phonon_displacements(
                displacements=random_rattle.output,  # pylint: disable=E1101
                structure=structure,
                supercell_matrix=None,
                phonon_maker=self.displacement_maker,
            )

            jobs.append(vasp_random_displacement_calcs)
            outputs.append(vasp_random_displacement_calcs.output["dirs"])

        # create a flow including all jobs
        return Flow(jobs=jobs, output=outputs, name=self.name)


@dataclass
class MLPhononMaker(FFPhononMaker):
    """
    Maker to calculate harmonic phonons with a force field.

    Calculate the harmonic phonons of a material. Initially, a tight structural
    relaxation is performed to obtain a structure without forces on the atoms.
    Subsequently, supercells with one displaced atom are generated and accurate
    forces are computed for these structures. With the help of phonopy, these
    forces are then converted into a dynamical matrix. To correct for polarization
    effects, a correction of the dynamical matrix based on BORN charges can
    be performed. The BORN charges can be supplied manually.
    Finally, phonon densities of states, phonon band structures
    and thermodynamic properties are computed.

    Notes
    -----
        It is heavily recommended to symmetrize the structure before passing it to
        this flow. Otherwise, a different space group might be detected and too
        many displacement calculations will be generated.
        It is recommended to check the convergence parameters here and
        adjust them if necessary. The default might not be strict enough
        for your specific case.

    Parameters
    ----------
    name : str
        Name of the flows produced by this maker.
    sym_reduce : bool
        Whether to reduce the number of deformations using symmetry.
    symprec : float
        Symmetry precision to use in the
        reduction of symmetry to find the primitive/conventional cell
        (use_primitive_standard_structure, use_conventional_standard_structure)
        and to handle all symmetry-related tasks in phonopy
    displacement: float
        Displacement distance for phonons
    min_length: float
        Minimum length of the supercell that will be built
    prefer_90_degrees: bool
        If set to True, supercell algorithm will first try to find a supercell
        with 3 90 degree angles
    get_supercell_size_kwargs: dict
        Keyword arguments that will be passed to get_supercell_size to determine supercell size
    use_symmetrized_structure: str
        Allowed strings: "primitive", "conventional", None

        - "primitive" will enforce to start the phonon computation
          from the primitive standard structure
          according to Setyawan, W., & Curtarolo, S. (2010).
          High-throughput electronic band structure calculations:
          Challenges and tools. Computational Materials Science,
          49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010.
          This makes it possible to use certain k-path definitions
          with this workflow. Otherwise, we must rely on seekpath
        - "conventional" will enforce to start the phonon computation
          from the conventional standard structure
          according to Setyawan, W., & Curtarolo, S. (2010).
          High-throughput electronic band structure calculations:
          Challenges and tools. Computational Materials Science,
          49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010.
          We will, however, use seekpath and primitive structures
          as determined by from phonopy to compute the phonon band structure
    bulk_relax_maker: .ForceFieldRelaxMaker or None
        Maker to perform a tight relaxation on the bulk.
        Set to ``None`` to skip the
        bulk relaxation
    static_energy_maker: .ForceFieldStaticMaker or None
        Maker to perform the computation of the DFT energy on the bulk.
        Set to ``None`` to skip the
        static energy computation
    phonon_displacement_maker: .ForceFieldStaticMaker or None
        Maker used to compute the forces for a supercell.
    generate_frequencies_eigenvectors_kwargs : dict
        Keyword arguments passed to :obj:`generate_frequencies_eigenvectors`.
    create_thermal_displacements: bool
        Arg that determines if thermal_displacement_matrices are computed
    kpath_scheme: str
        Scheme to generate kpoints. Please be aware that
        you can only use seekpath with any kind of cell
        Otherwise, please use the standard primitive structure
        Available schemes are:
        "seekpath", "hinuma", "setyawan_curtarolo", "latimer_munro".
        "seekpath" and "hinuma" are the same definition but
        seekpath can be used with any kind of unit cell as
        it relies on phonopy to handle the relationship
        to the primitive cell and not pymatgen
    code: str
        Determines the DFT code. currently only vasp is implemented.
        This keyword might enable the implementation of other codes
        in the future
    store_force_constants: bool
        If True, force constants will be stored
    relax_maker_kwargs: dict
        Keyword arguments that can be passed to the RelaxMaker.
    static_maker_kwargs: dict
        Keyword arguments that can be passed to the StaticMaker.
    """

    name: str = "ml phonon"
    min_length: float | None = 20.0
    displacement: float = 0.01
    bulk_relax_maker: ForceFieldRelaxMaker | None = field(
        default_factory=lambda: ForceFieldRelaxMaker(
            relax_cell=True,
            relax_kwargs={"interval": 500},
            force_field_name="GAP",
        )
    )
    phonon_displacement_maker: ForceFieldStaticMaker | None = field(
        default_factory=lambda: ForceFieldStaticMaker(
            name="gap phonon static",
            force_field_name="GAP",
        )
    )
    static_energy_maker: ForceFieldStaticMaker | None = None
    store_force_constants: bool = False
    get_supercell_size_kwargs: dict = field(
        default_factory=lambda: {"max_atoms": 20000, "step_size": 0.1}
    )
    generate_frequencies_eigenvectors_kwargs: dict = field(
        default_factory=lambda: {"units": "THz", "tol_imaginary_modes": 1e-1}
    )
    relax_maker_kwargs: dict | None = field(default_factory=dict)
    static_maker_kwargs: dict | None = field(default_factory=dict)

    @job
    def make_from_ml_model(
        self,
        structure,
        potential_file,
        ml_model: str = "GAP",
        calculator_kwargs: dict | None = None,
        supercell_settings: dict | None = None,
        **make_kwargs,
    ):
        """
        Maker for GAP phonon jobs.

        Parameters
        ----------
        structure : .Structure
            The pymatgen structure. Please start with a structure
            that is nearly fully optimized as the internal optimizers
            have very strict settings!
        ml_model: str
            ML model to be used. Default is GAP.
        potential_file :
            Complete path to MLIP file(s) (train, test and MLIP files)
        calculator_kwargs :
            Keyword arguments for the ASE Calculator.
        supercell_settings:
            Dict with supercell settings.
        make_kwargs :
            Keyword arguments for the PhononMaker.

        Returns
        -------
        PhononMaker jobs.

        """
        if supercell_settings is None:
            supercell_settings = field(default_factory=lambda: {"min_length": 15})
        if ml_model == "GAP":
            if calculator_kwargs is None:
                calculator_kwargs = {
                    "args_str": "IP GAP",
                    "param_filename": str(potential_file),
                }

            ml_prep = ml_phonon_maker_preparation(
                bulk_relax_maker=self.bulk_relax_maker,
                phonon_displacement_maker=self.phonon_displacement_maker,
                static_energy_maker=self.static_energy_maker,
                calculator_kwargs=calculator_kwargs,
                relax_maker_kwargs=self.relax_maker_kwargs,
                static_maker_kwargs=self.static_maker_kwargs,
            )

        elif ml_model == "J-ACE":
            raise UserWarning("No atomate2 ACE.jl PhononMaker implemented.")

        elif ml_model == "NEP":
            if calculator_kwargs is None:
                calculator_kwargs = {"model_filename": str(potential_file)}

            ml_prep = ml_phonon_maker_preparation(
                bulk_relax_maker=ForceFieldRelaxMaker(
                    relax_cell=True,
                    relax_kwargs={"interval": 500},
                    force_field_name="NEP",
                ),
                phonon_displacement_maker=ForceFieldStaticMaker(
                    name="nep phonon static",
                    force_field_name="NEP",
                ),
                static_energy_maker=None,
                calculator_kwargs=calculator_kwargs,
                relax_maker_kwargs=self.relax_maker_kwargs,
                static_maker_kwargs=self.static_maker_kwargs,
            )
        elif ml_model == "NEQUIP":
            if calculator_kwargs is None:
                calculator_kwargs = {
                    "model_path": str(potential_file),
                    "device": "cuda",
                }
            else:
                calculator_kwargs.update({"model_path": str(potential_file)})

            ml_prep = ml_phonon_maker_preparation(
                bulk_relax_maker=ForceFieldRelaxMaker(
                    relax_cell=True,
                    relax_kwargs={"interval": 500},
                    force_field_name="Nequip",
                ),
                phonon_displacement_maker=ForceFieldStaticMaker(
                    name="nequip phonon static",
                    force_field_name="Nequip",
                ),
                static_energy_maker=None,
                calculator_kwargs=calculator_kwargs,
                relax_maker_kwargs=self.relax_maker_kwargs,
                static_maker_kwargs=self.static_maker_kwargs,
            )

        elif ml_model == "M3GNET":
            if calculator_kwargs is None:
                calculator_kwargs = {"path": str(potential_file)}

            ml_prep = ml_phonon_maker_preparation(
                bulk_relax_maker=ForceFieldRelaxMaker(
                    relax_cell=True,
                    relax_kwargs={"interval": 500},
                    force_field_name="M3GNet",
                ),
                phonon_displacement_maker=ForceFieldStaticMaker(
                    name="m3gnet phonon static",
                    force_field_name="M3GNet",
                ),
                static_energy_maker=None,
                calculator_kwargs=calculator_kwargs,
                relax_maker_kwargs=self.relax_maker_kwargs,
                static_maker_kwargs=self.static_maker_kwargs,
            )

        else:  # MACE
            if calculator_kwargs is None:
                calculator_kwargs = {"model": str(potential_file), "device": "cuda"}
            elif "model" in calculator_kwargs:
                calculator_kwargs.update(
                    {"default_dtype": "float64"}
                )  # Use float64 for geometry optimization.
            else:
                calculator_kwargs.update(
                    {"model": str(potential_file), "default_dtype": "float64"}
                )

            ml_prep = ml_phonon_maker_preparation(
                bulk_relax_maker=ForceFieldRelaxMaker(
                    relax_cell=True,
                    relax_kwargs={"interval": 500},
                    force_field_name="MACE",
                ),
                phonon_displacement_maker=ForceFieldStaticMaker(
                    name="mace phonon static",
                    force_field_name="MACE",
                ),
                static_energy_maker=None,
                calculator_kwargs=calculator_kwargs,
                relax_maker_kwargs=self.relax_maker_kwargs,
                static_maker_kwargs=self.static_maker_kwargs,
            )

        (
            self.bulk_relax_maker,
            self.phonon_displacement_maker,
            self.static_energy_maker,
        ) = ml_prep
        filtered_settings = {
            key: value
            for key, value in supercell_settings.items()
            if key
            in [
                "min_length",
                "max_length",
                "fallback_min_length",
                "max_atoms",
                "min_atoms",
                "step_size",
            ]
        }
        supercell_matrix = reduce_supercell_size(
            structure=structure, **filtered_settings
        )

        flow = self.make(
            structure=structure, supercell_matrix=supercell_matrix, **make_kwargs
        )
        return Response(replace=flow, output=flow.output)


@dataclass
class IsoAtomStaticMaker(StaticMaker):
    """
    Maker to create Isolated atoms static (VASP) jobs.

    Parameters
    ----------
    name : str
        The job name.
    input_set_generator : .VaspInputGenerator
        Generator used to make the input set.
    write_input_set_kwargs : dict
        Keyword arguments that will get passed to :obj:`.write_vasp_input_set`.
    copy_vasp_kwargs : dict
        Keyword arguments that will get passed to :obj:`.copy_vasp_outputs`.
    run_vasp_kwargs : dict
        Keyword arguments that will get passed to :obj:`.run_vasp`.
    task_document_kwargs : dict
        Keyword arguments that will get passed to :obj:`.TaskDoc.from_directory`.
    stop_children_kwargs : dict
        Keyword arguments that will get passed to :obj:`.should_stop_children`.
    write_additional_data : dict
        Additional data to write to the current directory. Given as a dict of
        {filename: data}. Note that if using FireWorks, dictionary keys cannot contain
        the "." character which is typically used to denote file extensions. To avoid
        this, use the ":" character, which will automatically be converted to ".". E.g.
        ``{"my_file:txt": "contents of the file"}``.
    """

    name: str = "static"
    input_set_generator: VaspInputGenerator = field(
        default_factory=lambda: StaticSetGenerator(
            user_kpoints_settings={"reciprocal_density": 1},
            user_incar_settings={
                "ALGO": "Normal",
                "ISPIN": 1,
                "LAECHG": False,
                "ISMEAR": 0,
                "LCHARG": False,  # Do not write the CHGCAR file
                "LWAVE": False,  # Do not write the WAVECAR file
                "LVTOT": False,  # Do not write LOCPOT file
                "LORBIT": None,  # No output of projected or partial DOS in EIGENVAL, PROCAR and DOSCAR
                "LOPTICS": False,  # No PCDAT file
                # to be removed
                "NPAR": 4,
            },
        )
    )


@dataclass
class IsoAtomMaker(Maker):
    """
    Maker to generate DFT data for ML potential fitting from isolated atoms.

    Parameters
    ----------
    name : str
        Name of the flows produced by this maker.
    """

    name: str = "IsolatedAtomEnergyMaker"

    def make(
        self,
        all_species: list[Species],
        isolated_atom_maker: IsoAtomStaticMaker = None,
    ):
        """
        Make a flow to calculate the isolated atom's energy.

        Parameters
        ----------
        all_species : list[Species]
            List of pymatgen specie object.
        isolated_atom_maker: IsoAtomMaker
            VASP input set for the isolated atom calculation.
        """
        jobs = []
        isoatoms_energy = []
        isoatoms_dirs = []
        if isolated_atom_maker is None:
            isolated_atom_static_input_set = StaticSetGenerator(
                user_kpoints_settings={"grid_density": 1},
                user_incar_settings={
                    "ALGO": "Normal",
                    "ISPIN": 1,
                    "LAECHG": False,
                    "ISMEAR": 0,
                    "LCHARG": False,  # Do not write the CHGCAR file
                    "LWAVE": False,  # Do not write the WAVECAR file
                    "LVTOT": False,  # Do not write LOCPOT file
                    "LORBIT": None,  # No output of projected or partial DOS in EIGENVAL, PROCAR and DOSCAR
                    "LOPTICS": False,  # No PCDAT file
                    # to be removed
                    "NPAR": 4,
                    # TODO: locpot, chgcar, chg can be deactivated!
                    # TODO: why don't we use the IsoAtomMaker and adapt it?
                },
            )
            isolated_atom_maker = StaticMaker(
                input_set_generator=isolated_atom_static_input_set, name="stat_iso_atom"
            )
        for species in all_species:
            site = Site(species=species, coords=[0, 0, 0])
            mol = Molecule.from_sites([site])
            iso_atom = mol.get_boxed_structure(a=20, b=20, c=20)
            isolated_atom_maker.name = f"{species}-stat_iso_atom"
            isolated_atom_maker.run_vasp_kwargs = {"handlers": ()}
            isoatom_calcs = isolated_atom_maker.make(iso_atom)

            jobs.append(isoatom_calcs)
            isoatoms_energy.append(isoatom_calcs.output.output.energy_per_atom)
            isoatoms_dirs.append(isoatom_calcs.output.dir_name)

        # create a flow including all jobs
        return Flow(
            jobs=jobs,
            output={"energies": isoatoms_energy, "dirs": isoatoms_dirs},
            name=self.name,
        )



"""Flows to benchmark ML potentials."""

from dataclasses import dataclass

from atomate2.common.schemas.phonons import PhononBSDOSDoc
from jobflow import Maker, job
from pymatgen.core.structure import Structure

from autoplex.benchmark.phonons.utils import compute_bandstructure_benchmark_metrics

__all__ = ["PhononBenchmarkMaker"]


@dataclass
class PhononBenchmarkMaker(Maker):
    """
    Maker to benchmark all chosen ML potentials on the DFT (VASP) reference data.

    Produces a phonon band structure comparison and q-point-wise phonons RMSE plots,
    as well as a summary text file.

    Parameters
    ----------
    name : str
        Name of the flow produced by this maker.
    """

    name: str = "PhononBenchmark"

    @job
    def make(
        self,
        ml_model: str,
        structure: Structure,
        benchmark_mp_id: str,
        ml_phonon_task_doc: PhononBSDOSDoc,
        dft_phonon_task_doc: PhononBSDOSDoc,
        displacement: float,
        atomwise_regularization_parameter: float,
        soap_dict: dict,
        suffix: str,
    ):
        """
        Make flow for benchmarking.

        Parameters
        ----------
        ml_model: str
            ML model to be used. Default is GAP.
        structure:
            The pymatgen structures drawn from the Materials Project.
        benchmark_mp_id: str.
            Materials Project IDs for the structure.
        ml_phonon_task_doc: PhononBSDOSDoc
            Phonon task doc from ML potential consisting of pymatgen band-structure object.
        dft_phonon_task_doc: PhononBSDOSDoc
            Phonon task doc from DFT runs consisting of pymatgen band-structure object.
        displacement: float
            Displacement for finite displacement method.
        atomwise_regularization_parameter: float
            Regularization value for the atom-wise force components.
        suffix: str
            GAP potential file path suffix.
        soap_dict: dict
            Dictionary containing SOAP parameters.

        """
        return compute_bandstructure_benchmark_metrics(
            ml_model=ml_model,
            ml_phonon_bs=ml_phonon_task_doc.phonon_bandstructure,
            dft_phonon_bs=dft_phonon_task_doc.phonon_bandstructure,
            dft_imag_modes=dft_phonon_task_doc.has_imaginary_modes,
            ml_imag_modes=ml_phonon_task_doc.has_imaginary_modes,
            structure=structure,
            displacement=displacement,
            atomwise_regularization_parameter=atomwise_regularization_parameter,
            soap_dict=soap_dict,
            suffix=suffix,
            mp_id=benchmark_mp_id,
        )



"""Jobs for running RSS."""

import os
import re
from dataclasses import dataclass
from multiprocessing import Pool
from pathlib import Path
from shutil import which
from subprocess import run
from typing import Literal

import ase.io
import numpy as np
from ase import Atoms
from ase.data import atomic_numbers, covalent_radii
from jobflow import Flow, Maker, Response, job
from monty.dev import requires
from pymatgen.core import Element, Structure
from pymatgen.io.ase import AseAtomsAdaptor

from autoplex.data.common.utils import flatten
from autoplex.data.rss.utils import minimize_structures, split_structure_into_groups


@dataclass
class RandomizedStructure(Maker):
    """
    Maker to create random structures using the 'buildcell' tool.

    Parameters
    ----------
    name: str
        Name of the flows produced by this maker.
    struct_number : int
        Expected number of generated randomized unit cells.
    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
        passed into the parameters of buildcell. However, note that this will be overwritten
        if the stoichiometric ratio of elements is defined in the 'cell_seed_paths' or 'buildcell_options'.
    output_file_name: str
        Name of the file to store all generated structures.
    remove_tmp_files: bool
        Remove all temporary files raised by buildcell to save memory.
    buildcell_option: dict
        Customized parameters for buildcell.
    cell_seed_path: str
        Path to the custom buildcell control file, which ends with '.cell'. If this file exists,
        the buildcell_option argument will no longer take effect.
    num_processes: int
        Number of processes to use for parallel computation.
    fragment: Atoms | list[Atoms] (optional)
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] (optional)
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    """

    name: str = "build_random_cells"
    struct_number: int = 20
    tag: str = "Si"
    output_file_name: str = "random_structs.extxyz"
    remove_tmp_files: bool = True
    buildcell_option: dict | None = None
    cell_seed_path: str | None = None
    num_processes: int = 32
    fragment_file: str | None = None
    fragment_numbers: list[str] | None = None

    @requires(
        which("buildcell"),
        "RSS flows requires the executable 'buildcell' to be in PATH. "
        "Please follow the instructions in the autoplex documentation to install "
        "the AIRSS library and add it to PATH. Link to the documentation:"
        " https://autoatml.github.io/autoplex/user/index.html#enabling-rss-workflows",
    )
    @job
    def make(self):
        """Maker to create random structures by buildcell."""
        if self.cell_seed_path:
            if not os.path.isfile(self.cell_seed_path):
                raise FileNotFoundError(
                    f"No file found at the specified path: {self.cell_seed_path}"
                )
            bc_file = self.cell_seed_path

        else:
            buildcell_parameters = [
                "SLACK=0.25",
                "OVERLAP=0.1",
                "COMPACT",
                "MINSEP=1.5",
            ]

            if self.buildcell_option is not None:
                buildcell_parameters = self._update_buildcell_option(
                    self.buildcell_option, buildcell_parameters
                )

            elements = self._extract_elements(self.tag)  # {"Si":1, "O":2}

            if (
                self.buildcell_option is not None
                and "SPECIES" in self.buildcell_option
                and self.fragment_file is not None
            ):
                raise ValueError(
                    "Cannot use 'SPECIES' and 'fragment' together in buildcell options.\n"
                    "Specify your fragment only and use NFORM to control their number."
                )

            if self.buildcell_option is None or (
                "SPECIES" not in self.buildcell_option and self.fragment_file is None
            ):
                make_species = self._make_species(elements)  # Si%NUM=1,O%NUM=2
                buildcell_parameters = self._update_buildcell_option(
                    {"SPECIES": make_species}, buildcell_parameters
                )

            if (
                self.buildcell_option is None
                or (
                    "VARVOL" not in self.buildcell_option
                    and "TARGVOL" not in self.buildcell_option
                )
                or "MINSEP" not in self.buildcell_option
            ):
                r0 = {}
                varvol = {}
                num_atom_formula = 0
                total_varvol_formula = 0

                for ele in elements:
                    r0[ele] = covalent_radii[atomic_numbers[ele]]

                    if Element(ele).is_metal:
                        varvol[ele] = 5.5 * np.power(r0[ele], 3)
                    else:
                        varvol[ele] = 14.5 * np.power(r0[ele], 3)

                    total_varvol_formula += varvol[ele] * elements[ele]

                    num_atom_formula += elements[ele]

                if self.buildcell_option is None or (
                    "VARVOL" not in self.buildcell_option
                    and "TARGVOL" not in self.buildcell_option
                ):
                    mean_var = total_varvol_formula / num_atom_formula * len(elements)
                    buildcell_parameters = self._update_buildcell_option(
                        {"TARGVOL": f"{mean_var*0.8}-{mean_var*1.2}"},
                        buildcell_parameters,
                    )

                if (
                    self.buildcell_option is None
                    or "MINSEP" not in self.buildcell_option
                ):
                    minsep = self._make_minsep(r0)
                    buildcell_parameters = self._update_buildcell_option(
                        {
                            "MINSEP": minsep,
                        },
                        buildcell_parameters,
                    )

            if self.fragment_file is not None:
                self.fragment = ase.io.read(self.fragment_file, index=":")

                if len(self.fragment) == 1:
                    self.fragment = self.fragment[0]

                if isinstance(self.fragment, Atoms):
                    if self.fragment_numbers is None:
                        fragment_numbers = [1 for _ in self.fragment]
                    else:
                        fragment_numbers = self.fragment_numbers
                    if "fragment_id" not in self.fragment.arrays:
                        self.fragment.arrays["fragment_id"] = [
                            f"{1}-f" for i in self.fragment
                        ]

                elif isinstance(self.fragment, list):
                    if self.fragment_numbers is None:
                        fragment_numbers = [
                            1 for _ in range(sum([len(i) for i in self.fragment]))
                        ]
                    else:
                        fragment_numbers = self.fragment_numbers
                    write_fragment = self.fragment[0]
                    for frag in self.fragment[
                        1:
                    ]:  # merge all separate fragments into one Atoms object
                        write_fragment += frag

                fragment_parameters = [
                    "%BLOCK POSITIONS_ABS",
                ]
                symbols = self.fragment.get_chemical_symbols()
                for i, val in enumerate(self.fragment.get_positions(wrap=True)):
                    if i == 0:
                        newline = (
                            f"{symbols[i]} {val[0]:.8f} {val[1]:.8f} {val[2]:.8f}"
                            f" # {self.fragment.arrays['fragment_id'][i]}"
                            f" % NUM={fragment_numbers[i]}"
                        )
                    else:
                        newline = (
                            f"{symbols[i]} {val[0]:.8f} {val[1]:.8f} {val[2]:.8f}"
                            f" # {self.fragment.arrays['fragment_id'][i]}"
                        )
                    fragment_parameters.append(newline)
                fragment_parameters.append("%ENDBLOCK POSITIONS_ABS")

                buildcell_parameters = (
                    fragment_parameters + buildcell_parameters
                )  # prepend with structural info

            self._cell_seed(buildcell_parameters, self.tag)
            bc_file = f"{self.tag}.cell"

        with Pool(processes=self.num_processes) as pool:
            args = [
                (i, bc_file, self.tag, self.remove_tmp_files)
                for i in range(self.struct_number)
            ]
            atoms_group = pool.starmap(self._parallel_process, args)

        atoms_group = [
            atom for atom in atoms_group if not np.isnan(atom.get_positions()).any()
        ]

        ase.io.write(
            self.output_file_name, atoms_group, parallel=False, format="extxyz"
        )

        # structure = [AseAtomsAdaptor().get_structure(at) for at in atoms_group]
        return os.path.join(Path.cwd(), self.output_file_name)
        # return structure

    def _update_buildcell_option(self, updates, origin) -> list:
        """
        Update buildcell parameters based on a dictionary of updates.

        Parameters
        ----------
        updates: dict
            A dictionary consisting of new values to update buildcell parameters.
        origin: list
            The default list of buildcell parameters.
        """
        updated_keys = set()

        for i, option in enumerate(origin):
            option_key = option.split("=")[0]
            if option_key in updates:
                origin[i] = f"{option_key}={updates[option_key]}"
                updated_keys.add(option_key)

        for key, value in updates.items():
            if key not in updated_keys:
                origin.append(f"{key}={value}")

        return origin

    def _cell_seed(
        self,
        buildcell_parameters: list,
        tag: str,
    ):
        """
        Prepare the seed file for buildcell.

        Parameters
        ----------
        buildcell_parameters: (list of str) e.g. ['VARVOL=20']
            List of parameters for creating the seed file for buildcell.
        tag: str
            Tag of systems.
        """
        bc_file = f"{tag}.cell"
        contents = []
        flag = False  # for printing blocks correctly with '#'
        for i in buildcell_parameters:
            if i.startswith("%") or flag:
                flag = not (flag and i.startswith("%"))
                contents.append(i + "\n")
            else:
                contents.append("#" + i + "\n")

        with open(bc_file, "w") as f:
            f.writelines(contents)

    def _extract_elements(self, input_str: str) -> dict[str, int]:
        """
        Extract elements and their counts from a chemical formula string.

        Parameters
        ----------
        input_str: str
            A string representing a chemical formula (e.g., "SiO2").

        Returns
        -------
        Dict[str, int]
            A dictionary. For example, the input "SiO2" would return {"Si": 1, "O": 2}.
        """
        elements: dict[str, int] = {}
        pattern = re.compile(r"([A-Z][a-z]*)(\d*)")
        matches = pattern.findall(input_str)

        for match in matches:
            element, count = match
            count = int(count) if count else 1
            if element in elements:
                elements[element] += count
            else:
                elements[element] = count

        return elements

    def _make_species(self, elements: dict[str, int]) -> str:
        """
        Create a formatted string from a dictionary of element symbols and their counts.

        Parameters
        ----------
        elements: dict
            A dictionary of element symbols and their counts, e.g., {"Si": 1, "O": 2}.

        Returns
        -------
        str
            A formatter string. For example, the input {"Si": 1, "O": 2} would return "Si%NUM=1,O%NUM=2".
        """
        output = ""
        for element, count in elements.items():
            output += f"{element}%NUM={count},"
        return output[:-1]

    def _make_minsep(self, r: dict[str, float]) -> str:
        """
        Generate a minsep string based on the radii of the elements.

        Parameters
        ----------
        r: dict
            A dictionary of element symbols and their atomic radii.

        Returns
        -------
        str
            A formatted string. For example, the input {"Si": 1.1, "O": 0.66} would
            return "1.5 Si-Si=1.76 Si-O=1.408 O-O=1.056".

        TODO: set up robust heuristics for multi-component systems
        """
        keys = list(r.keys())
        if len(keys) == 1:
            return str(1.6 * r[keys[0]])

        minsep = "1.5 "
        for i in range(len(keys)):
            for j in range(i, len(keys)):
                el1, el2 = keys[i], keys[j]
                r1, r2 = r[el1], r[el2]
                result = (r1 + r2) / 2 * 1.6

                minsep += f"{el1}-{el2}={result} "

        return minsep[:-1]

    def _parallel_process(
        self, i: int, bc_file: str, tag: str, remove_tmp_files: bool
    ) -> Atoms:
        """
        Run the 'buildcell' command in parallel.

        Parameters
        ----------
        i: int
            Unique index to differentiate temporary files.
        bc_file: str
            Path to the input 'buildcell' file.
        tag: str
            Tag used to differentiate temporary files.
        remove_tmp_files: bool
            If True, remove temporary files after processing.

        """
        tmp_file_name = "tmp." + str(i) + "." + tag + ".cell"
        tmp_error_file_name = "tmp_error." + str(i) + "." + tag + ".cell"

        with (
            open(bc_file) as bc_file_handle,
            open(tmp_file_name, "w") as tmp_file_handle,
            open(tmp_error_file_name, "w") as tmp_error_file_handle,
        ):
            run(
                "buildcell",
                stdin=bc_file_handle,
                stdout=tmp_file_handle,
                stderr=tmp_error_file_handle,
                shell=True,
                check=True,
            )

        atom = ase.io.read(tmp_file_name, parallel=False)
        atom.info["unique_starting_index"] = i

        if "castep_labels" in atom.arrays:
            del atom.arrays["castep_labels"]

        if "initial_magmoms" in atom.arrays:
            del atom.arrays["initial_magmoms"]

        if remove_tmp_files:
            os.remove(tmp_file_name)
            os.remove(tmp_error_file_name)

        return atom


@job
def do_rss_single_node(
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"],
    mlip_path: str | Path,
    iteration_index: str,
    structures: list[Structure],
    output_file_name: str = "RSS_relax_results",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 1000,
    force_tol: float = 0.01,
    stress_tol: float = 0.01,
    hookean_repul: bool = False,
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None = None,
    num_processes_rss: int = 1,
    device: str = "cpu",
    isolated_atom_energies: dict[int, float] | None = None,
    struct_start_index: int = 0,
    config_type: str = "traj",
    keep_symmetry: bool = True,
) -> list[str | None]:
    """
    Perform sandom structure searching (RSS) on one node using a machine learning interatomic potential (MLIP).

    Parameters
    ----------
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"]
        Choose one specific MLIP type to be fitted.
    mlip_path: str | Path
        Path to the MLIP model.
    iteration_index: str
        Index for the current iteration.
    structures: list[Structure]
        List of structures to be relaxed.
    output_file_name: str
        Prefix for the trajectory/log file name. The actual output file name
        may be composed of this prefix, an index, and file types.
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 1000.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.01.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.01.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    num_processes_rss: int
        Number of processes used for running RSS.
    device: str
        Specify device to use "cuda" or "cpu".
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    struct_start_index: int
        Specify the starting index within a list
    config_type: str
        Specify the type of configurations generated from RSS
    keep_symmetry: bool
        If true, preserve symmetry during relaxation.

    Returns
    -------
    list
        Output list[str] containing paths for the results of the RSS relaxation.
    """
    return minimize_structures(
        mlip_type=mlip_type,
        mlip_path=mlip_path,
        iteration_index=iteration_index,
        structures=structures,
        output_file_name=output_file_name,
        scalar_pressure_method=scalar_pressure_method,
        scalar_exp_pressure=scalar_exp_pressure,
        scalar_pressure_exponential_width=scalar_pressure_exponential_width,
        scalar_pressure_low=scalar_pressure_low,
        scalar_pressure_high=scalar_pressure_high,
        max_steps=max_steps,
        force_tol=force_tol,
        stress_tol=stress_tol,
        hookean_repul=hookean_repul,
        hookean_paras=hookean_paras,
        num_processes_rss=num_processes_rss,
        device=device,
        isolated_atom_energies=isolated_atom_energies,
        struct_start_index=struct_start_index,
        config_type=config_type,
        keep_symmetry=keep_symmetry,
    )


@job
def do_rss_multi_node(
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"],
    mlip_path: str | Path,
    iteration_index: str,
    structure: list[Structure] | list[list[Structure]] | None = None,
    structure_paths: str | list[str] | None = None,
    output_file_name: str = "RSS_relax_results",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 1000,
    force_tol: float = 0.01,
    stress_tol: float = 0.01,
    hookean_repul: bool = False,
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None = None,
    num_processes_rss: int = 1,
    device: str = "cpu",
    isolated_atom_energies: dict[int, float] | None = None,
    num_groups: int = 1,
    config_type: str = "traj",
    keep_symmetry: bool = True,
) -> list[list | None]:
    """
    Perform sandom structure searching (RSS) on multiple nodes using a machine learning interatomic potential (MLIP).

    Parameters
    ----------
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"]
        Choose one specific MLIP type to be fitted.
    mlip_path: str | Path
        Path to the MLIP model.
    iteration_index: str
        Index for the current iteration.
    structure: list[Structure]
        List of structures to be relaxed.
    structure_paths: str | list[str]
        Path(s) to structures to be used in the RSS process.
    output_file_name: str
        Prefix for the trajectory/log file name. The actual output file name
        may be composed of this prefix, an index, and file types.
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 1000.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.01.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.01.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    num_processes_rss: int
        Number of processes used for running RSS.
    device: str
        Specify device to use "cuda" or "cpu".
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    num_groups: int
        Number of structure groups, used for assigning tasks across multiple nodes,
        with each node handling one group.
    config_type: str
        Specify the type of configurations generated from RSS
    keep_symmetry: bool
        If true, preserve symmetry during relaxation.

    Returns
    -------
    list
        Output list[str] containing paths for the results of the RSS relaxation.
    """
    job_list = []

    if structure_paths is not None:
        if isinstance(structure_paths, list):
            atoms = [ase.io.read(dir, index=":") for dir in structure_paths]
            atoms = flatten(atoms, recursive=True)
        elif isinstance(structure_paths, str):
            atoms = ase.io.read(structure_paths, index=":")
        structure = [AseAtomsAdaptor().get_structure(at) for at in atoms]
    elif structure is not None and isinstance(structure[0], list):
        structure = flatten(structure, recursive=False)

    if structure is not None and isinstance(structure, list):
        structure_groups = split_structure_into_groups(structure, num_groups)
    else:
        raise ValueError("Invalid structure format. It must be a list of structures.")

    rss_info = []

    struct_start_index = 0

    for i in range(num_groups):
        rss = do_rss_single_node(
            mlip_type=mlip_type,
            mlip_path=mlip_path,
            iteration_index=iteration_index,
            structures=structure_groups[i],
            output_file_name=output_file_name,
            scalar_pressure_method=scalar_pressure_method,
            scalar_exp_pressure=scalar_exp_pressure,
            scalar_pressure_exponential_width=scalar_pressure_exponential_width,
            scalar_pressure_low=scalar_pressure_low,
            scalar_pressure_high=scalar_pressure_high,
            max_steps=max_steps,
            force_tol=force_tol,
            stress_tol=stress_tol,
            hookean_repul=hookean_repul,
            hookean_paras=hookean_paras,
            num_processes_rss=num_processes_rss,
            device=device,
            isolated_atom_energies=isolated_atom_energies,
            struct_start_index=struct_start_index,
            config_type=config_type,
            keep_symmetry=keep_symmetry,
        )

        struct_start_index += len(structure_groups[i])

        job_list.append(rss)
        rss_info.append(rss.output)

    return Response(replace=Flow(job_list), output=rss_info)



"""Utilities for testing castep calculations.

License / Copyright
-------------------
This code is adapted from `atomate2/src/atomate2/utils/testing/aims.py`.

The code has been released under BSD 3-Clause License
and the following copyright applies:

atomate2 Copyright (c) 2015, The Regents of the University of
California, through Lawrence Berkeley National Laboratory (subject
to receipt of any required approvals from the U.S. Dept. of Energy).
All rights reserved.


Reference:

Ganose, A. M., Sahasrabuddhe, H., Asta, M., Beck, K., Biswas, T., Bonkowski, A.,
Bustamante, J., Chen, X., Chiang, Y., Chrzan, D. C., Clary, J., Cohen, O. A.,
Ertural, C., Gallant, M. C., George, J., Gerits, S., Goodall, R. E. A.,
Guha, R. D., Hautier, G., Horton, M., Inizan, T. J., Kaplan, A. D., Kingsbury, R. S.,
Kuner, M. C., Li, B., Linn, X., McDermott, M. J., Mohanakrishnan, R. S., Naik, A. A.,
Neaton, J. B., Parmar, S. M., Persson, K. A., Petretto, G., Purcell, T. A. R.,
Ricci, F., Rich, B., Riebesell, J., Rignanese, G.-M., Rosen, A. S., Scheffler, M.,
Schmidt, J., Shen, J.-X., Sobolev, A., Sundararaman, R., Tezak, C., Trinquet, V.,
Varley, J. B., Vigil-Fowler, D., Wang, D., Waroquiers, D., Wen, M., Yang, H.,
Zheng, H., Zheng, J., Zhu, Z., & Jain, A. (2025). Atomate2: modular workflows
for materials science. Digital Discovery. DOI: 10.1039/D5DD00019J.
"""

from __future__ import annotations

import logging
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, Literal

from jobflow import CURRENT_JOB
from monty.os.path import zpath as monty_zpath

import autoplex.misc.castep.jobs
from autoplex.misc.castep.utils import CastepStaticSetGenerator

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Sequence

    from pytest import MonkeyPatch

    from autoplex.misc.castep.utils import CastepInputGenerator

logger = logging.getLogger("autoplex")


_VFILES: Final = ("castep.param", "castep.cell")
_REF_PATHS: dict[str, str | Path] = {}
_FAKE_RUN_CASTEP_KWARGS: dict[str, dict] = {}


def zpath(path: str | Path) -> Path:
    """Return the path of a zip file.

    Returns an existing (zipped or unzipped) file path given the unzipped
    version. If no path exists, returns the unmodified path.
    """
    return Path(monty_zpath(str(path)))


def monkeypatch_castep(
    monkeypatch: MonkeyPatch, ref_path: Path
) -> Generator[Callable[[Any, Any], Any]]:
    """Fixture to mock CASTEP runs for tests.

    Usage
    -----
    1. "mock_castep" should be included as an argument to any test that would like to use
       its functionally.
    2. For each job in your workflow, you should prepare a reference directory
       containing two folders "inputs" (containing the reference input files expected
       to be produced by Castep().initialize()) and "outputs" (containing the expected
       output files to be produced by run_castep). These files should reside in a
       subdirectory of "tests/test_data/castep".
    3. Create a dictionary mapping each job name to its reference directory. Note that
       you should supply the reference directory relative to the "tests/test_data/castep"
       folder. For example, if your calculation has one job named "static" and the
       reference files are present in "tests/test_data/castep/Si_static", the dictionary
       would look like: ``{"static": "Si_static"}``.
    4. Inside the test function, call `mock_castep(ref_paths, fake_castep_kwargs)`, where
       ref_paths is the dictionary created in step 3.
    5. Run your castep job after calling `mock_castep`.
    For examples, see the tests in tests/misc/castep/test_jobs.py.

    Parameters
    ----------
    monkeypatch
        MonkeyPatch fixture used to set attributes.
    ref_path : Path
        Path to the root folder containing the ``tests/test_data/castep`` reference data.

    Returns
    -------
    Generator[Callable[..., None], None, None]
        A generator that yields a helper function which accepts the reference
        mapping and optional kwargs. After the test, the fixture undoes the
        monkeypatches and clears the internal mappings.
    """

    def mock_run_castep(*args, **kwargs) -> None:
        name = CURRENT_JOB.job.name
        try:
            ref_dir = ref_path / _REF_PATHS[name]
        except KeyError:
            raise ValueError(
                f"no reference directory found for job {name!r}; "
                f"reference paths received={_REF_PATHS}"
            ) from None
        fake_run_castep(ref_dir, **_FAKE_RUN_CASTEP_KWARGS.get(name, {}))

    get_input_set_orig = CastepStaticSetGenerator.get_input_set

    def mock_get_input_set(
        self: CastepStaticSetGenerator, *args, **kwargs
    ) -> CastepInputGenerator:
        return get_input_set_orig(self, *args, **kwargs)

    monkeypatch.setattr(autoplex.misc.castep.run, "run_castep", mock_run_castep)
    monkeypatch.setattr(autoplex.misc.castep.jobs, "run_castep", mock_run_castep)
    monkeypatch.setattr(CastepStaticSetGenerator, "get_input_set", mock_get_input_set)

    def _run(ref_paths: dict, fake_run_castep_kwargs: dict | None = None) -> None:
        _REF_PATHS.update(ref_paths)
        _FAKE_RUN_CASTEP_KWARGS.update(fake_run_castep_kwargs or {})

    yield _run

    monkeypatch.undo()
    _REF_PATHS.clear()
    _FAKE_RUN_CASTEP_KWARGS.clear()


def fake_run_castep(
    ref_path: str | Path,
    input_settings: Sequence[str] | None = None,
    check_inputs: Sequence[Literal["castep.param", "castep.cell"]] = _VFILES,
    clear_inputs: bool = False,
) -> None:
    """Emulate running castep and validate castep input files.

    Parameters
    ----------
    ref_path
        Path to reference directory with castep input files in the folder named 'inputs'
        and output files in the folder named 'outputs'.
    input_settings
        A list of input settings to check.
    check_inputs
        A list of castep input files to check. Supported options are "castep.inp"
    clear_inputs
        Whether to clear input files before copying in the reference castep outputs.
    """
    logger.info("Running fake castep.")

    ref_path = Path(ref_path)

    logger.info("Verified inputs successfully")

    if clear_inputs:
        clear_castep_inputs()

    copy_castep_outputs(ref_path)

    # pretend to run castep by copying pre-generated outputs from reference dir
    logger.info("Generated fake castep outputs")


def clear_castep_inputs() -> None:
    """Clean up castep input files."""
    for castep_file in ("castep.param", "castep.cell"):
        if Path(castep_file).exists():
            Path(castep_file).unlink()
    logger.info("Cleared castep inputs")


def copy_castep_outputs(ref_path: str | Path) -> None:
    """Copy castep output files from the reference directory.

    Parameters
    ----------
    ref_path : str or Path
        Path to the reference directory containing an ``outputs`` subfolder
        with the CASTEP output files to copy.
    """
    output_path = Path(ref_path) / "outputs"
    for output_file in output_path.iterdir():
        if output_file.is_file():
            shutil.copy(output_file, "./CASTEP")



"""General fitting jobs using several MLIPs available."""

from pathlib import Path

import numpy as np
from ase.io import write
from jobflow import job
from pymatgen.io.ase import AseAtomsAdaptor

from autoplex import MLIP_HYPERS
from autoplex.fitting.common.utils import (
    check_convergence,
    gap_fitting,
    jace_fitting,
    m3gnet_fitting,
    mace_fitting,
    nep_fitting,
    nequip_fitting,
)


@job
def machine_learning_fit(
    database_dir: str | Path,
    species_list: list,
    run_fits_on_different_cluster: bool = False,
    isolated_atom_energies: dict | None = None,
    num_processes_fit: int = 32,
    auto_delta: bool = True,
    glue_xml: bool = False,
    glue_file_path: str = "glue.xml",
    gpu_identifier_indices: list[int] | None = None,
    mlip_type: str | None = None,
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    device: str = "cuda",
    database_dict: dict | None = None,
    hyperpara_opt: bool = False,
    hyperparameters: MLIP_HYPERS = MLIP_HYPERS,
    **fit_kwargs,
):
    """
    Job for fitting potential(s).

    Parameters
    ----------
    database_dir: Str | Path
        Path to the directory containing the database.
    species_list: list
        List of element names (strings) involved in the training dataset
    run_fits_on_different_cluster: bool
        Whether to run fitting on different clusters.
    isolated_atom_energies: dict
        Dictionary of isolated atoms energies.
    num_processes_fit: int
        Number of processes for fitting.
    auto_delta: bool
        Automatically determine delta for 2b, 3b and soap terms. Only used for GAP fitting.
    glue_xml: bool
        Use the glue.xml core potential instead of fitting 2b terms. Only used for GAP fitting.
    glue_file_path: str
        Name of the glue.xml file path. Only used for GAP fitting.
    gpu_identifier_indices: list[int]
        List of GPU indices to be used for fitting. Only used for NEP fitting.
    mlip_type: str
        Choose one specific MLIP type to be fitted:
        'GAP' | 'J-ACE' | 'NEQUIP' | 'NEP' | 'M3GNET' | 'MACE'
    ref_energy_name: str
        Reference energy name.
    ref_force_name: str
        Reference force name.
    ref_virial_name: str
        Reference virial name.
    device: str
        Device to be used for model fitting, either "cpu" or "cuda".
    database_dict: dict
        Dict including all training and test databases.
    hyperpara_opt: bool
        Perform hyperparameter optimization using XPOT
        (XPOT: https://pubs.aip.org/aip/jcp/article/159/2/024803/2901815)
    hyperparameters: MLIP_HYPERS
        Hyperparameters for MLIP fitting.
    run_fits_on_different_cluster: bool
        Indicates if fits are to be run on a different cluster.
        If True, the fitting data (train.extxyz, test.extxyz) is stored in the database.
    fit_kwargs: dict
        Additional keyword arguments for MLIP fitting.
    """
    if run_fits_on_different_cluster:

        adapter = AseAtomsAdaptor()
        for key, values in database_dict.items():
            if values is not None:
                if not Path(key).parent.exists():
                    Path(key).parent.mkdir(parents=True, exist_ok=True)

                for value in values:
                    properties = value.properties.copy()
                    properties["REF_virial"] = np.array(properties["REF_virial"])
                    value.properties = properties
                    new_value = adapter.get_atoms(value)
                    write(key, new_value, parallel=False, format="extxyz", append=True)
        database_dir = Path().cwd()

    else:
        if isinstance(database_dir, str):  # data_prep_job.output is returned as string
            database_dir = Path(database_dir)

    train_files = [
        "train.extxyz",
        "without_regularization/train.extxyz",
        "phonon/train.extxyz",
        "rattled/train.extxyz",
    ]
    test_files = [
        "test.extxyz",
        "without_regularization/test.extxyz",
        "phonon/test.extxyz",
        "rattled/test.extxyz",
    ]

    mlip_paths = []

    if mlip_type == "GAP":
        for train_name, test_name in zip(train_files, test_files):
            if (database_dir / train_name).exists() and (
                database_dir / test_name
            ).exists():
                train_test_error = gap_fitting(
                    db_dir=database_dir,
                    hyperparameters=hyperparameters.GAP,
                    species_list=species_list,
                    num_processes_fit=num_processes_fit,
                    auto_delta=auto_delta,
                    glue_xml=glue_xml,
                    glue_file_path=glue_file_path,
                    ref_energy_name=ref_energy_name,
                    ref_force_name=ref_force_name,
                    ref_virial_name=ref_virial_name,
                    train_name=train_name,
                    test_name=test_name,
                    fit_kwargs=fit_kwargs,
                )
                mlip_paths.append(train_test_error["mlip_path"])

    elif mlip_type == "J-ACE":
        train_test_error = jace_fitting(
            db_dir=database_dir,
            hyperparameters=hyperparameters.J_ACE,
            isolated_atom_energies=isolated_atom_energies,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            num_processes_fit=num_processes_fit,
            fit_kwargs=fit_kwargs,
        )
        mlip_paths.append(train_test_error["mlip_path"])

    elif mlip_type == "NEP":
        if gpu_identifier_indices is None:
            gpu_identifier_indices = [0]

        train_test_error = nep_fitting(
            db_dir=database_dir,
            hyperparameters=hyperparameters.NEP,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            species_list=species_list,
            gpu_identifier_indices=gpu_identifier_indices,
            fit_kwargs=fit_kwargs,
        )

        mlip_paths.append(train_test_error["mlip_path"])

    elif mlip_type == "NEQUIP":
        train_test_error = nequip_fitting(
            db_dir=database_dir,
            hyperparameters=hyperparameters.NEQUIP,
            isolated_atom_energies=isolated_atom_energies,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            fit_kwargs=fit_kwargs,
            device=device,
        )
        mlip_paths.append(train_test_error["mlip_path"])

    elif mlip_type == "M3GNET":
        train_test_error = m3gnet_fitting(
            db_dir=database_dir,
            hyperparameters=hyperparameters.M3GNET,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            fit_kwargs=fit_kwargs,
            device=device,
        )
        mlip_paths.append(train_test_error["mlip_path"])

    elif mlip_type == "MACE":
        train_test_error = mace_fitting(
            db_dir=database_dir,
            hyperparameters=hyperparameters.MACE,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            device=device,
            fit_kwargs=fit_kwargs,
        )
        mlip_paths.append(train_test_error["mlip_path"])

    check_conv = check_convergence(train_test_error["test_error"])

    return {
        "mlip_path": mlip_paths,
        "train_error": train_test_error["train_error"],
        "test_error": train_test_error["test_error"],
        "convergence": check_conv,
        "database_dir": database_dir,
    }



"""Jobs to create training data for ML potentials."""

from jobflow import job
from pymatgen.core.structure import Structure

from autoplex.data.phonons.utils import reduce_supercell_size


@job
def reduce_supercell_size_job(
    structure: Structure,
    min_length: float = 18,
    max_length: float = 20,
    fallback_min_length: float = 12,
    min_atoms: int = 100,
    max_atoms: int = 500,
    step_size: float = 1,
):
    """
    Reduce phonopy supercell size.

    Parameters
    ----------
    structure: Structure
        The pymatgen Structure object.
    min_length: float
        Minimum length of the supercell that will be built.
    max_length: float
        Maximum length of the supercell that will be built.
    max_atoms: int
        Maximally allowed number of atoms in the supercell.
    min_atoms: int
        Minimum number of atoms in the supercell that shall be reached.
    fallback_min_length: float
        Fallback option for minimum length for exceptional cases.
    step_size: float
        The step_size which is used to increase the supercell.
        If allow_orthorhombic and force_90_degrees are both set to True,
        the chosen step_size will be automatically multiplied by 5 to
        prevent a too long search for the possible supercell.

    Returns
    -------
    reduced_supercell_size call.
    """
    return reduce_supercell_size(
        structure=structure,
        min_length=min_length,
        max_length=max_length,
        fallback_min_length=fallback_min_length,
        min_atoms=min_atoms,
        max_atoms=max_atoms,
        step_size=step_size,
    )



"""Flows for running RSS."""

from dataclasses import dataclass

from jobflow import Flow, Maker, Response, job

from autoplex.data.common.jobs import (
    sample_data,
)
from autoplex.data.rss.jobs import RandomizedStructure

__all__ = ["BuildMultiRandomizedStructure"]


@dataclass
class BuildMultiRandomizedStructure(Maker):
    """
    Maker to create random structures by 'buildcell'.

    Parameters
    ----------
    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
        passed into the parameters of buildcell. However, note that this will be overwritten
        if the stoichiometric ratio of elements is defined in the 'cell_seed_paths' or 'buildcell_options'.
    generated_struct_numbers: list[int]
        Expected number of generated randomized unit cells.
    cell_seed_paths: list[str]
        A list of paths to the custom buildcell control files, which ends with '.cell'. If these files exist,
        the buildcell_options argument will no longer take effect.
    buildcell_options: list[dict]
        A list of customized buildcell parameters.
    fragment_file: Atoms | list[Atoms] (optional)
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] (optional)
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    remove_tmp_files: bool
        Remove all temporary files raised by buildcell to save memory.
    initial_selection_enabled: bool
        If true, sample structures using CUR.
    selected_struct_numbers: list
        Number of structures to be sampled.
    bcur_params: dict
        Parameters for Boltzmann CUR selection.
    random_seed: int
        A seed to ensure reproducibility of CUR selection.
    num_processes: int
        Number of processes to use for parallel computation.
    name: str
        Name of the flows produced by this maker.

    """

    tag: str
    generated_struct_numbers: list[int]
    cell_seed_paths: list[str] | None = None
    buildcell_options: list[dict] | None = None
    fragment_file: str | None = None
    fragment_numbers: list[str] | None = None
    remove_tmp_files: bool = True
    initial_selection_enabled: bool = False
    selected_struct_numbers: list[int] | None = None
    bcur_params: dict | None = None
    random_seed: int | None = None
    num_processes: int = 1
    name: str = "do_randomized_structure_generation"

    @job
    def make(self):
        """Maker to create random structures by buildcell."""
        job_list = []
        final_structures = []
        for i, struct_number in enumerate(self.generated_struct_numbers):
            cell_seed_path = None
            buildcell_option = None
            if self.cell_seed_paths is not None:
                assert len(self.generated_struct_numbers) == len(self.cell_seed_paths)
                cell_seed_path = self.cell_seed_paths[i]
            elif self.buildcell_options is not None:
                assert len(self.generated_struct_numbers) == len(self.buildcell_options)
                buildcell_option = self.buildcell_options[i]
            job_struct = RandomizedStructure(
                tag=self.tag,
                struct_number=struct_number,
                remove_tmp_files=self.remove_tmp_files,
                cell_seed_path=cell_seed_path,
                buildcell_option=buildcell_option,
                fragment_file=self.fragment_file,
                fragment_numbers=self.fragment_numbers,
                num_processes=self.num_processes,
            ).make()
            job_struct.name = f"{self.name}_{i}"

            if self.initial_selection_enabled:
                assert len(self.generated_struct_numbers) == len(
                    self.selected_struct_numbers
                )
                job_cur = sample_data(
                    selection_method="cur",
                    num_of_selection=self.selected_struct_numbers[i],
                    bcur_params=self.bcur_params,
                    dir=job_struct.output,
                    random_seed=self.random_seed,
                )
                job_cur.name = f"sampling_{i}"
                job_list.append(job_struct)
                job_list.append(job_cur)
                final_structures.append(job_cur.output)
            else:
                job_list.append(job_struct)
                final_structures.append(job_struct.output)

        return Response(
            replace=Flow(job_list),
            output=final_structures,
        )



"""
AutoPLEX -- Automated machine-learned Potential Landscape explorer -- flows.

This package is designed to perform a workflow-based
automated machine-learned interatomic potentials (MLIP) fit for
DFT-labelled data.

"""

from autoplex._version import __version__
from autoplex.settings import (
    GAPSettings,
    JACESettings,
    M3GNETSettings,
    MACESettings,
    MLIPHypers,
    NEPSettings,
    NEQUIPSettings,
    RssConfig,
)

MLIP_HYPERS = MLIPHypers()
RSS_CONFIG = RssConfig()
GAP_HYPERS = GAPSettings()
JACE_HYPERS = JACESettings()
M3GNET_HYPERS = M3GNETSettings()
MACE_HYPERS = MACESettings()
NEQUIP_HYPERS = NEQUIPSettings()
NEP_HYPERS = NEPSettings()



"""RSS Jobs include the generation of the initial potential model as well as iterative RSS exploration."""

import logging
from typing import Literal

from atomate2.forcefields.jobs import ForceFieldStaticMaker
from atomate2.vasp.jobs.base import BaseVaspMaker
from atomate2.vasp.jobs.core import StaticMaker
from atomate2.vasp.sets.core import StaticSetGenerator
from jobflow import Flow, Response, job

from autoplex.data.common.flows import DFTStaticLabelling
from autoplex.data.common.jobs import (
    collect_dft_data,
    preprocess_data,
    sample_data,
)
from autoplex.data.rss.flows import BuildMultiRandomizedStructure
from autoplex.data.rss.jobs import do_rss_multi_node
from autoplex.fitting.common.flows import MLIPFitMaker
from autoplex.misc.castep.jobs import CastepStaticMaker

__all__ = ["do_rss_iterations", "initial_rss"]

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

_DEFAULT_STATIC_ENERGY_MAKER = StaticMaker(
    input_set_generator=StaticSetGenerator(
        user_incar_settings={
            "ADDGRID": "True",
            "ENCUT": 520,
            "EDIFF": 1e-06,
            "ISMEAR": 0,
            "SIGMA": 0.01,
            "PREC": "Accurate",
            "ISYM": None,
            "KSPACING": 0.2,
            "NPAR": 8,
            "LWAVE": "False",
            "LCHARG": "False",
            "ENAUG": None,
            "GGA": None,
            "ISPIN": None,
            "LAECHG": None,
            "LELF": None,
            "LORBIT": None,
            "LVTOT": None,
            "NSW": None,
            "SYMPREC": None,
            "NELM": 100,
            "LMAXMIX": None,
            "LASPH": None,
            "AMIN": None,
        }
    ),
    run_vasp_kwargs={"handlers": ()},
)


@job
def initial_rss(
    tag: str,
    generated_struct_numbers: list[int],
    num_of_initial_selected_structs: list[int] | None = None,
    cell_seed_paths: list[str] | None = None,
    buildcell_options: list[dict] | None = None,
    fragment_file: str | None = None,
    fragment_numbers: list[str] | None = None,
    num_processes_buildcell: int = 1,
    initial_selection_enabled: bool = False,
    bcur_params: dict | None = None,
    random_seed: int | None = None,
    include_isolated_atom: bool = False,
    isolatedatom_box: list[float] | None = None,
    e0_spin: bool = False,
    include_dimer: bool = False,
    dimer_box: list[float] | None = None,
    dimer_range: list | None = None,
    dimer_num: int = 21,
    custom_incar: dict | None = None,
    custom_potcar: dict | None = None,
    config_type: str | None = None,
    dft_ref_file: str = "dft_ref.extxyz",
    rss_group: str = "initial",
    test_ratio: float = 0.1,
    regularization: bool = False,
    retain_existing_sigma: bool = False,
    scheme: str | None = None,
    element_order: list | None = None,
    reg_minmax: list[tuple] | None = None,
    distillation: bool = False,
    force_max: float | None = None,
    force_label: str = "REF_forces",
    pre_database_dir: str | None = None,
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"] = "GAP",
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    auto_delta: bool = False,
    num_processes_fit: int = 1,
    device_for_fitting: str = "cpu",
    static_energy_maker: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
    ) = _DEFAULT_STATIC_ENERGY_MAKER,
    static_energy_maker_isolated_atoms: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker | None
    ) = None,
    **fit_kwargs,
):
    """
    Run initial Random Structure Searching (RSS) workflow from scratch.

    Parameters
    ----------
    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
        passed into the parameters of buildcell. However, note that this will be overwritten
        if the stoichiometric ratio of elements is defined in the 'cell_seed_paths' or 'buildcell_options'.
    generated_struct_numbers: list[int]
        Expected number of generated randomized unit cells.
    num_of_initial_selected_structs: list[int] | None
        Number of structures to be sampled. Default is None.
    cell_seed_paths: list[str]
        A list of paths to the custom buildcell control files, which ends with '.cell'. If these files exist,
        the buildcell_options argument will no longer take effect.
    buildcell_options: list[dict] | None
        Customized parameters for buildcell. Default is None.
    fragment_file: Atoms | list[Atoms] | None
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] | None
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    num_processes_buildcell: int
        Number of processes to use for parallel computation during buildcell generation. Default is 1.
    initial_selection_enabled: bool
        If true, sample structures using CUR. Default is False.
    bcur_params: dict | None
        Parameters for Boltzmann CUR selection. Default is None.
    random_seed: int | None
        A seed to ensure reproducibility of CUR selection. Default is None.
    include_isolated_atom: bool
        If true, perform single-point calculations for isolated atoms. Default is False.
    isolatedatom_box: list[float] | None
        List of the lattice constants for an isolated atom configuration. Default is None.
    e0_spin: bool
        If true, include spin polarization in isolated atom and dimer calculations. Default is False.
    include_dimer: bool
        If true, perform single-point calculations for dimers. Default is False.
    dimer_box: list[float] | None
        The lattice constants of a dimer box. Default is None.
    dimer_range: list[float] | None
        Range of distances for dimer calculations. Default is None.
    dimer_num: int
        Number of different distances to consider for dimer calculations. Default is 21.
    custom_incar: dict | None
        Dictionary of custom VASP input parameters. If provided, will update the
        default parameters. Default is None.
    custom_potcar: dict | None
        Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
        Default is None.
    config_type: str | None
        Configuration type for the DFT calculations. Default is None.
    dft_ref_file: str
        Reference file for DFT-labelled data. Default is 'dft_ref.extxyz'.
    rss_group: str
        Group name for GAP RSS. Default is 'initial'.
    test_ratio: float
        The proportion of the test set after splitting the data.
        If None, no splitting will be performed. Default is 0.1.
    regularization: bool
        If true, apply regularization. This only works for GAP. Default is False.
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.
    scheme: str | None
        Scheme to use for regularization. Default is None.
    element_order:
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2).
        This value is useful when constructing high-dimensional convex hulls based on the
        "volume-stoichiometry" scheme. Specially, if the dataset contains compounds with
        different numbers of constituent elements (e.g., both binary and ternary structures),
        this value must be explicitly set to ensure the convex hull is constructed consistently.
    reg_minmax: list[tuple] | None
        A list of tuples representing the minimum and maximum values for regularization.
    distillation: bool
        If true, apply data distillation. Default is False.
    force_max: float | None
        Maximum force value to exclude structures. Default is None.
    force_label: str
        The label of force values to use for distillation. Default is 'REF_forces'.
    pre_database_dir: str | None
        Directory where the previous database was saved. Default is None.
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"]
        Choose one specific MLIP type to be fitted. Default is 'GAP'.
    ref_energy_name: str
        Reference energy name. Default is 'REF_energy'.
    ref_force_name: str
        Reference force name. Default is 'REF_forces'.
    ref_virial_name: str
        Reference virial name. Default is 'REF_virial'.
    auto_delta: bool
        If true, apply automatic determination of delta for GAP terms. Default is False.
    num_processes_fit: int
        Number of processes used for fitting. Default is 1.
    device_for_fitting: str
        Device to be used for model fitting, either "cpu" or "cuda".
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs: either BaseVaspMaker (VASP-based) or CastepStaticMaker
        (CASTEP-based) or ForceFieldStaticMaker (force field-based). Defaults to StaticMaker (VASP-based).
    static_energy_maker_isolated_atoms: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs of isolated atoms: either BaseVaspMaker (VASP-based) or CastepStaticMaker
        (CASTEP-based) or ForceFieldStaticMaker (force field-based) or None. If set to `None`, the parameters
        from `static_energy_maker` will be used as the default for isolated atoms. In this case,
        if `static_energy_maker` is a `StaticMaker`, all major settings will be inherited,
        except that `kspacing` will be automatically set to 100 to enforce a Gamma-point-only calculation.
        This is typically suitable for single-atom systems. Default is None. If a non-`StaticMaker` maker
        is used here, its output must include a `dir_name` field to ensure compatibility with downstream workflows.
    fit_kwargs:
        Additional keyword arguments for the MLIP fitting process.

    Returns
    -------
    dict:
        A dictionary with following information

        - 'test_error': float, The test error of the fitted MLIP.
        - 'pre_database_dir': str, The directory of the preprocessed database.
        - 'mlip_path': Path to the fitted MLIP.
        - 'isolated_atom_energies': dict, The isolated energy values.
        - 'current_iter': int, The current iteration index, set to 0.
    """
    if isolatedatom_box is None:
        isolatedatom_box = [20.0, 20.0, 20.0]
    if dimer_box is None:
        dimer_box = [20.0, 20.0, 20.0]

    do_randomized_structure_generation = BuildMultiRandomizedStructure(
        generated_struct_numbers=generated_struct_numbers,
        cell_seed_paths=cell_seed_paths,
        buildcell_options=buildcell_options,
        fragment_file=fragment_file,
        fragment_numbers=fragment_numbers,
        selected_struct_numbers=num_of_initial_selected_structs,
        tag=tag,
        num_processes=num_processes_buildcell,
        initial_selection_enabled=initial_selection_enabled,
        bcur_params=bcur_params,
        random_seed=random_seed,
    ).make()

    # TODO: this needs to be generalized beyond VASP and instead be able to use a different dft calculator,
    # or a force field
    do_dft_static = DFTStaticLabelling(
        e0_spin=e0_spin,
        isolatedatom_box=isolatedatom_box,
        isolated_atom=include_isolated_atom,
        dimer=include_dimer,
        dimer_box=dimer_box,
        dimer_range=dimer_range,
        dimer_num=dimer_num,
        custom_incar=custom_incar,
        custom_potcar=custom_potcar,
        static_energy_maker=static_energy_maker,
        static_energy_maker_isolated_atoms=static_energy_maker_isolated_atoms,
    ).make(
        structures=do_randomized_structure_generation.output, config_type=config_type
    )
    do_data_collection = collect_dft_data(
        dft_ref_file=dft_ref_file, rss_group=rss_group, dft_dirs=do_dft_static.output
    )
    do_data_preprocessing = preprocess_data(
        test_ratio=test_ratio,
        regularization=regularization,
        retain_existing_sigma=retain_existing_sigma,
        scheme=scheme,
        element_order=element_order,
        distillation=distillation,
        force_max=force_max,
        force_label=force_label,
        dft_ref_dir=do_data_collection.output["dft_ref_dir"],
        pre_database_dir=pre_database_dir,
        reg_minmax=reg_minmax,
        isolated_atom_energies=do_data_collection.output["isolated_atom_energies"],
    )
    do_mlip_fit = MLIPFitMaker(
        mlip_type=mlip_type,
        ref_energy_name=ref_energy_name,
        ref_force_name=ref_force_name,
        ref_virial_name=ref_virial_name,
        num_processes_fit=num_processes_fit,
        apply_data_preprocessing=False,
        auto_delta=auto_delta,
        glue_xml=False,
    ).make(
        isolated_atom_energies=do_data_collection.output["isolated_atom_energies"],
        database_dir=do_data_preprocessing.output,
        device=device_for_fitting,
        **fit_kwargs,
    )

    job_list = [
        do_randomized_structure_generation,
        do_dft_static,
        do_data_collection,
        do_data_preprocessing,
        do_mlip_fit,
    ]

    return Response(
        replace=Flow(job_list),
        output={
            "test_error": do_mlip_fit.output["test_error"],
            "pre_database_dir": do_data_preprocessing.output,
            "mlip_path": do_mlip_fit.output["mlip_path"][0],
            "isolated_atom_energies": do_data_collection.output[
                "isolated_atom_energies"
            ],
        },
    )


@job
def do_rss_iterations(
    input: dict,
    tag: str,
    generated_struct_numbers: list[int],
    num_of_initial_selected_structs: list[int] | None = None,
    cell_seed_paths: list[str] | None = None,
    buildcell_options: list[dict] | None = None,
    fragment_file: str | None = None,
    fragment_numbers: list[str] | None = None,
    num_processes_buildcell: int = 1,
    initial_selection_enabled: bool = False,
    rss_selection_method: str = None,
    num_of_rss_selected_structs: int = 100,
    bcur_params: dict | None = None,
    random_seed: int | None = None,
    include_isolated_atom: bool = False,
    isolatedatom_box: list[float] | None = None,
    e0_spin: bool = False,
    include_dimer: bool = False,
    dimer_box: list[float] | None = None,
    dimer_range: list | None = None,
    dimer_num: int = 21,
    custom_incar: dict | None = None,
    custom_potcar: dict | None = None,
    config_types: list[str] | None = None,
    dft_ref_file: str = "dft_ref.extxyz",
    rss_group: str = "rss",
    test_ratio: float = 0.1,
    regularization: bool = False,
    retain_existing_sigma: bool = False,
    scheme: str | None = None,
    element_order: list | None = None,
    reg_minmax: list[tuple] | None = None,
    distillation: bool = True,
    force_max: float = 200,
    force_label: str = "REF_forces",
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"] = "GAP",
    ref_energy_name: str = "REF_energy",
    ref_force_name: str = "REF_forces",
    ref_virial_name: str = "REF_virial",
    auto_delta: bool = False,
    num_processes_fit: int = 1,
    device_for_fitting: str = "cpu",
    scalar_pressure_method: str = "exp",
    scalar_exp_pressure: float = 100,
    scalar_pressure_exponential_width: float = 0.2,
    scalar_pressure_low: float = 0,
    scalar_pressure_high: float = 50,
    max_steps: int = 200,
    force_tol: float = 0.05,
    stress_tol: float = 0.05,
    hookean_repul: bool = False,
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None = None,
    keep_symmetry: bool = False,
    remove_traj_files: bool = False,
    num_processes_rss: int = 1,
    device_for_rss: str = "cpu",
    stop_criterion: float = 0.01,
    max_iteration_number: int = 5,
    num_groups: int = 1,
    initial_kt: float = 0.3,
    current_iter_index: int = 1,
    static_energy_maker: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
    ) = _DEFAULT_STATIC_ENERGY_MAKER,
    static_energy_maker_isolated_atoms: (
        BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker | None
    ) = None,
    **fit_kwargs,
):
    """
    Perform iterative RSS to improve the accuracy of a MLIP.

    Each iteration involves generating new structures, sampling, running
    DFT calculations, collecting data, preprocessing data, and fitting a new MLIP.

    Parameters
    ----------
    input : dict
        A dictionary parameter used to pass specific input data required during the RSS iterations.
        The keys in this dictionary should be one of the following valid keys:

            test_error: float
                The test error of the fitted MLIP.
            pre_database_dir: str
                The directory of the preprocessed database.
            mlip_path: str | path
                Path to the fitted MLIP.
            isolated_atom_energies: dict
                The isolated energy values.
            current_iter: int
                The current iteration index.
            kt: float
                The value of kt.

    tag: str
        Tag of systems. It can also be used for setting up elements and stoichiometry.
        For example, the tag of 'SiO2' will be recognized as a 1:2 ratio of Si to O and
        passed into the parameters of buildcell. However, note that this will be overwritten
        if the stoichiometric ratio of elements is defined in the 'cell_seed_paths' or 'buildcell_options'.
    generated_struct_numbers: list[int]
        Expected number of generated randomized unit cells.
    num_of_initial_selected_structs: list[int] | None
        Number of structures to be sampled. Default is None.
    cell_seed_paths: list[str]
        A list of paths to the custom buildcell control files, which ends with '.cell'. If these files exist,
        the buildcell_options argument will no longer take effect.
    buildcell_options: list[dict] | None
        Customized parameters for buildcell. Default is None.
    fragment_file: Atoms | list[Atoms] | None
        Fragment(s) for random structures, e.g. molecules, to be placed indivudally intact.
        atoms.arrays should have a 'fragment_id' key with unique identifiers for each fragment if in same Atoms.
        atoms.cell must be defined (e.g. Atoms.cell = np.eye(3)*20).
    fragment_numbers: list[str] | None
        Numbers of each fragment to be included in the random structures. Defaults to 1 for all specified.
    num_processes_buildcell: int
        Number of processes to use for parallel computation during buildcell generation. Default is 1.
    initial_selection_enabled: bool
        If true, sample structures using CUR. Default is False.
    rss_selection_method: str
        Method for selecting samples from the generated structures. Default is None.
    num_of_rss_selected_structs: int
        Number of structures to be selected.
    bcur_params: dict | None
        Parameters for Boltzmann CUR selection. Default is None.
    random_seed: int | None
        A seed to ensure reproducibility of CUR selection. Default is None.
    include_isolated_atom: bool
        If true, perform single-point calculations for isolated atoms. Default is False.
    isolatedatom_box: list[float] | None
        List of the lattice constants for an isolated atom configuration. Default is None.
    e0_spin: bool
        If true, include spin polarization in isolated atom and dimer calculations. Default is False.
    include_dimer: bool
        If true, perform single-point calculations for dimers only once. Default is False.
    dimer_box: list[float] | None
        The lattice constants of a dimer box. Default is None.
    dimer_range: list[float] | None
        Range of distances for dimer calculations. Default is None.
    dimer_num: int
        Number of different distances to consider for dimer calculations. Default is 21.
    custom_incar: dict | None
        Dictionary of custom VASP input parameters. If provided, will update the
        default parameters. Default is None.
    custom_potcar: dict | None
        Dictionary of POTCAR settings to update. Keys are element symbols, values are the desired POTCAR labels.
        Default is None.
    config_types: list[str] | None
        Configuration types for the DFT calculations. Default is None.
    dft_ref_file: str
        Reference file for DFT-labelled data. Default is 'dft_ref.extxyz'.
    rss_group: str
        Group name for GAP RSS. Default is 'rss'.
    test_ratio: float
        The proportion of the test set after splitting the data. Default is 0.1.
    regularization: bool
        If true, apply regularization. This only works for GAP. Default is False.
    retain_existing_sigma: bool
        Whether to keep the current sigma values for specific configuration types.
        If set to True, existing sigma values for specific configurations will remain unchanged.
    scheme: str | None
        Scheme to use for regularization. Default is None.
    element_order:
        List of atomic numbers in order of choice (e.g. [42, 16] for MoS2).
        This value is useful when constructing high-dimensional convex hulls based on the
        "volume-stoichiometry" scheme. Specially, if the dataset contains compounds with
        different numbers of constituent elements (e.g., both binary and ternary structures),
        this value must be explicitly set to ensure the convex hull is constructed consistently.
    reg_minmax: list[tuple] | None
        A list of tuples representing the minimum and maximum values for regularization.
    distillation: bool
        If true, apply data distillation. Default is True.
    force_max: float
        Maximum force value to exclude structures. Default is 200.
    force_label: str
        The label of force values to use for distillation. Default is 'REF_forces'.
    mlip_type: Literal["GAP", "J-ACE", "NEP", "NEQUIP", "M3GNET", "MACE"]
        Choose one specific MLIP type to be fitted. Default is 'GAP'.
    ref_energy_name: str
        Reference energy name. Default is 'REF_energy'.
    ref_force_name: str
        Reference force name. Default is 'REF_forces'.
    ref_virial_name: str
        Reference virial name. Default is 'REF_virial'.
    auto_delta: bool
        If true, apply automatic determination of delta for GAP terms. Default is False.
    num_processes_fit: int
        Number of processes used for fitting. Default is 1.
    device_for_fitting: str
        Device to be used for model fitting, either "cpu" or "cuda".
    scalar_pressure_method: str
        Method for adding external pressures. Default is 'exp'.
    scalar_exp_pressure: float
        Scalar exponential pressure. Default is 100.
    scalar_pressure_exponential_width: float
        Width for scalar pressure exponential. Default is 0.2.
    scalar_pressure_low: float
        Low limit for scalar pressure. Default is 0.
    scalar_pressure_high: float
        High limit for scalar pressure. Default is 50.
    max_steps: int
        Maximum number of steps for relaxation. Default is 200.
    force_tol: float
        Force residual tolerance for relaxation. Default is 0.05.
    stress_tol: float
        Stress residual tolerance for relaxation. Default is 0.05.
    hookean_repul: bool
        If true, apply Hookean repulsion. Default is False.
    hookean_paras: dict[tuple[int, int], tuple[float, float]] | None
        Parameters for Hookean repulsion as a dictionary of tuples. Default is None.
    keep_symmetry: bool
        If true, preserve symmetry during relaxation. Default is False.
    remove_traj_files: bool
            If true, remove all trajectory files raised by RSS to save memory. Default is False.
    num_processes_rss: int
        Number of processes used for running RSS. Default is 1.
    device_for_rss: str
        Specify device to use "cuda" or "cpu" for running RSS. Default is "cpu".
    stop_criterion: float
        Convergence criterion for stopping RSS iterations. Default is 0.01.
    max_iteration_number: int
        Maximum number of RSS iterations to perform. Default is 5.
    num_groups: int
        Number of structure groups, used for assigning tasks across multiple nodes. Default is 1.
    initial_kt: float
        Initial temperature (in eV) for Boltzmann sampling. Default is 0.3.
    current_iter_index: int
        Index for the current RSS iteration. Default is 1.
    static_energy_maker: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs: either BaseVaspMaker (VASP-based) or CastepStaticMaker
        (CASTEP-based) or ForceFieldStaticMaker (force field-based). Defaults to StaticMaker (VASP-based).
    static_energy_maker_isolated_atoms: BaseVaspMaker | CastepStaticMaker | ForceFieldStaticMaker
        Maker for static energy jobs of isolated atoms: either BaseVaspMaker (VASP-based) or CastepStaticMaker
        (CASTEP-based) or ForceFieldStaticMaker (force field-based) or None. If set to `None`, the parameters
        from `static_energy_maker` will be used as the default for isolated atoms. In this case,
        if `static_energy_maker` is a `StaticMaker`, all major settings will be inherited,
        except that `kspacing` will be automatically set to 100 to enforce a Gamma-point-only calculation.
        This is typically suitable for single-atom systems. Default is None. If a non-`StaticMaker` maker
        is used here, its output must include a `dir_name` field to ensure compatibility with downstream workflows.
    fit_kwargs:
        Additional keyword arguments for the MLIP fitting process.

    Returns
    -------
    dict:
        A dictionary with following information

        - 'test_error': float, The test error of the fitted MLIP.
        - 'pre_database_dir': str, The directory of the preprocessed database.
        - 'mlip_path': Path to the fitted MLIP.
        - 'isolated_atom_energies': dict, The isolated energy values.
        - 'current_iter': int, The current iteration index.
        - 'kt': float, The temperature (in eV) for Boltzmann sampling.
    """
    test_error = input.get("test_error")
    current_iter = input.get("current_iter", current_iter_index)
    current_kt = input.get("kt", initial_kt)

    config_type = (
        (config_types[0] if current_kt > 0.1 else config_types[-1])
        if config_types
        else None
    )

    if isolatedatom_box is None:
        isolatedatom_box = [20.0, 20.0, 20.0]
    if dimer_box is None:
        dimer_box = [20.0, 20.0, 20.0]

    logging.info(
        f"The configuration type of structures generated in the current iteration will be {config_type}!"
    )

    if (
        test_error is not None
        and test_error > stop_criterion
        and current_iter is not None
        and current_iter < max_iteration_number
    ):
        logging.info(f"Current kt: {current_kt}")
        logging.info(f"Current iter index: {current_iter}")
        logging.info(f"The error of {current_iter}th iteration: {test_error}")

        if bcur_params is None:
            bcur_params = {}
        bcur_params["kt"] = current_kt

        do_randomized_structure_generation = BuildMultiRandomizedStructure(
            generated_struct_numbers=generated_struct_numbers,
            cell_seed_paths=cell_seed_paths,
            buildcell_options=buildcell_options,
            fragment_file=fragment_file,
            fragment_numbers=fragment_numbers,
            selected_struct_numbers=num_of_initial_selected_structs,
            tag=tag,
            num_processes=num_processes_buildcell,
            initial_selection_enabled=initial_selection_enabled,
            bcur_params=bcur_params,
            random_seed=random_seed,
        ).make()
        do_rss = do_rss_multi_node(
            mlip_type=mlip_type,
            iteration_index=f"{current_iter}th",
            mlip_path=input["mlip_path"],
            structure_paths=do_randomized_structure_generation.output,
            scalar_pressure_method=scalar_pressure_method,
            scalar_exp_pressure=scalar_exp_pressure,
            scalar_pressure_exponential_width=scalar_pressure_exponential_width,
            scalar_pressure_low=scalar_pressure_low,
            scalar_pressure_high=scalar_pressure_high,
            max_steps=max_steps,
            force_tol=force_tol,
            stress_tol=stress_tol,
            hookean_repul=hookean_repul,
            hookean_paras=hookean_paras,
            keep_symmetry=keep_symmetry,
            num_processes_rss=num_processes_rss,
            device=device_for_rss,
            num_groups=num_groups,
            config_type=config_type,
        )
        do_data_sampling = sample_data(
            selection_method=rss_selection_method,
            num_of_selection=num_of_rss_selected_structs,
            bcur_params=bcur_params,
            traj_path=do_rss.output,
            random_seed=random_seed,
            isolated_atom_energies=input["isolated_atom_energies"],
            remove_traj_files=remove_traj_files,
        )
        do_dft_static = DFTStaticLabelling(
            e0_spin=e0_spin,
            isolatedatom_box=isolatedatom_box,
            isolated_atom=include_isolated_atom,
            dimer=include_dimer,
            dimer_box=dimer_box,
            dimer_range=dimer_range,
            dimer_num=dimer_num,
            custom_incar=custom_incar,
            custom_potcar=custom_potcar,
            static_energy_maker=static_energy_maker,
            static_energy_maker_isolated_atoms=static_energy_maker_isolated_atoms,
        ).make(structures=do_data_sampling.output, config_type=config_type)
        do_data_collection = collect_dft_data(
            dft_ref_file=dft_ref_file,
            rss_group=rss_group,
            dft_dirs=do_dft_static.output,
        )
        do_data_preprocessing = preprocess_data(
            test_ratio=test_ratio,
            regularization=regularization,
            retain_existing_sigma=retain_existing_sigma,
            scheme=scheme,
            element_order=element_order,
            distillation=distillation,
            force_max=force_max,
            force_label=force_label,
            dft_ref_dir=do_data_collection.output["dft_ref_dir"],
            pre_database_dir=input["pre_database_dir"],
            reg_minmax=reg_minmax,
            isolated_atom_energies=input["isolated_atom_energies"],
        )
        do_mlip_fit = MLIPFitMaker(
            mlip_type=mlip_type,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            num_processes_fit=num_processes_fit,
            apply_data_preprocessing=False,
            auto_delta=auto_delta,
            glue_xml=False,
        ).make(
            database_dir=do_data_preprocessing.output,
            isolated_atom_energies=input["isolated_atom_energies"],
            device=device_for_fitting,
            **fit_kwargs,
        )

        kt = current_kt - 0.1 if (current_kt - 0.1) > 0.1 else 0.1
        current_iter += 1
        if include_isolated_atom:
            include_isolated_atom = False
        if include_dimer:
            include_dimer = False

        do_iteration = do_rss_iterations(
            input={
                "test_error": do_mlip_fit.output["test_error"],
                "pre_database_dir": do_data_preprocessing.output,
                "mlip_path": do_mlip_fit.output["mlip_path"][0],
                "isolated_atom_energies": input["isolated_atom_energies"],
                "current_iter": current_iter,
                "kt": kt,
            },
            generated_struct_numbers=generated_struct_numbers,
            num_of_initial_selected_structs=num_of_initial_selected_structs,
            tag=tag,
            cell_seed_paths=cell_seed_paths,
            buildcell_options=buildcell_options,
            fragment_file=fragment_file,
            fragment_numbers=fragment_numbers,
            num_processes_buildcell=num_processes_buildcell,
            initial_selection_enabled=initial_selection_enabled,
            rss_selection_method=rss_selection_method,
            num_of_rss_selected_structs=num_of_rss_selected_structs,
            bcur_params=bcur_params,
            random_seed=random_seed,
            e0_spin=e0_spin,
            isolatedatom_box=isolatedatom_box,
            include_isolated_atom=include_isolated_atom,
            include_dimer=include_dimer,
            dimer_box=dimer_box,
            dimer_range=dimer_range,
            dimer_num=dimer_num,
            custom_incar=custom_incar,
            custom_potcar=custom_potcar,
            config_types=config_types,
            dft_ref_file=dft_ref_file,
            rss_group=rss_group,
            test_ratio=test_ratio,
            regularization=regularization,
            retain_existing_sigma=retain_existing_sigma,
            scheme=scheme,
            element_order=element_order,
            reg_minmax=reg_minmax,
            distillation=distillation,
            force_max=force_max,
            force_label=force_label,
            mlip_type=mlip_type,
            ref_energy_name=ref_energy_name,
            ref_force_name=ref_force_name,
            ref_virial_name=ref_virial_name,
            auto_delta=auto_delta,
            num_processes_fit=num_processes_fit,
            scalar_pressure_method=scalar_pressure_method,
            scalar_exp_pressure=scalar_exp_pressure,
            scalar_pressure_exponential_width=scalar_pressure_exponential_width,
            scalar_pressure_low=scalar_pressure_low,
            scalar_pressure_high=scalar_pressure_high,
            max_steps=max_steps,
            force_tol=force_tol,
            stress_tol=stress_tol,
            hookean_repul=hookean_repul,
            hookean_paras=hookean_paras,
            keep_symmetry=keep_symmetry,
            remove_traj_files=remove_traj_files,
            num_processes_rss=num_processes_rss,
            device_for_rss=device_for_rss,
            stop_criterion=stop_criterion,
            max_iteration_number=max_iteration_number,
            num_groups=num_groups,
            initial_kt=initial_kt,
            current_iter_index=current_iter_index,
            static_energy_maker=static_energy_maker,
            static_energy_maker_isolated_atoms=static_energy_maker_isolated_atoms,
            **fit_kwargs,
        )

        job_list = [
            do_randomized_structure_generation,
            do_rss,
            do_data_sampling,
            do_dft_static,
            do_data_collection,
            do_data_preprocessing,
            do_mlip_fit,
            do_iteration,
        ]

        return Response(detour=job_list, output=do_iteration.output)

    return input



# Description: This script averages the test durations from all the artifacts
import glob
import json
import subprocess
from collections import defaultdict
import os

# Function to collect test names using pytest
def collect_tests_with_pytest():
    # Run pytest with --collect-only to get the list of test cases
    result = subprocess.run(
        ["pytest", "--collect-only"],
        capture_output=True,
        text=True
    )

    collected_tests = defaultdict(list)

    # Parse the output to organize tests under their respective modules
    for line in result.stdout.splitlines():
        if line.startswith("tests/"):
            collected_tests[line] = 0

    return collected_tests


# Consolidate durations from existing artifacts
def consolidate_durations():
    durations = defaultdict(lambda: {'total_duration': 0, 'count': 0})

    # Iterate over all downloaded duration artifacts
    for folder in glob.glob("test-durations-*"):
        # The path to the duration file in each directory
        duration_file_path = os.path.join(folder, ".pytest-split-durations")

        if os.path.isfile(duration_file_path):
            with open(duration_file_path, "r") as f:
                data = json.load(f)
                for test, duration in data.items():
                    durations[test]['total_duration'] += duration
                    durations[test]['count'] += 1

    # Calculate the average duration for each test
    return {test: info['total_duration'] / info['count'] for test, info in durations.items()}

# Define the path to the consolidated durations file
CONSOLIDATED_FILE = "tests/test_data/.pytest-split-durations"


# Main script logic
def main():
    # Collect tests grouped by modules using pytest
    collected_tests = collect_tests_with_pytest()

    # Consolidate durations from artifacts
    consolidated_durations = consolidate_durations()

    
    # Merge and update with consolidated durations
    updated_durations = {}
    for test, duration in collected_tests.items():
        # Update average test durations and exclude test which not exists (can happen on renaming or removing tests) 
        if test in consolidated_durations:
            updated_durations[test] = consolidated_durations[test]
        
    # Load the existing durations file if it exists
    existing_durations = {}
    if os.path.isfile(CONSOLIDATED_FILE):
        with open(CONSOLIDATED_FILE, "r") as f:
            existing_durations = json.load(f)

    
    # Sort the keys to compare the tests in both dictionaries
    updated_durations_key = sorted(updated_durations.keys())
    existing_durations_key = sorted(existing_durations.keys())


    # Check if all keys in updated_durations are in existing_durations
    if updated_durations_key == existing_durations_key:
        print("No new tests detected; durations file remains unchanged.")
    else:
        # Write the updated durations to the consolidated file
        with open(CONSOLIDATED_FILE, "w") as f:
            json.dump(updated_durations, f, indent=4)
        print("New tests detected; updated the durations file.")

if __name__ == "__main__":
    main()
